"""OpenCV functions for pyorc."""

import copy
import os
import warnings

import cv2
import numpy as np
import rasterio
from scipy import optimize
from shapely.affinity import rotate
from shapely.geometry import LineString, Point, Polygon
from tqdm import tqdm

from . import helpers

# default distortion coefficients for no distortion
DIST_COEFFS = [[0.0], [0.0], [0.0], [0.0], [0.0]]

# criteria for finding subpix corners
criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.001)


def _check_valid_frames(cap, frame_number):
    """Determine last valid index in the frame sequence of a video capture object.

    Determines where at the end of the sequence a frame cannot be retrieved due to some error.
    The function iterates backward through the frame numbers and sets the frame position to each one, attempting to
    read the frame. If a frame cannot be read, it updates the `last_valid_idx`
    to that position. If all frames are invalid, it defaults the index to zero.

    Parameters
    ----------
    cap : cv2.VideoCapture
        capture object.
    frame_number : list[int]
        A list representing the sequence of frame numbers to check for validity.

    Returns
    -------
    int or None
        The last valid index in the frame sequence or `None` if no invalid frames
        are found at the end of the sequence.

    """
    # start with None
    last_valid_idx = None
    ret = False
    n = -1
    while ret is False:
        cap.set(cv2.CAP_PROP_POS_FRAMES, frame_number[n])
        ret, img = cap.read()
        if not (ret):
            last_valid_idx = n
        n -= 1
        if n == -len(frame_number) - 1:
            last_valid_idx = 0
            break

    return last_valid_idx


def _combine_m(m1, m2):
    """Combine two 2x3 serial transformation matrices into one.

    Done by extending them to 3x3 matrices for matrix multiplication, compute the product, and return the resulting
    2x3 matrix.

    Parameters
    ----------
    m1 : array_like
        First matrix to be combined, should be of shape (2, 3).

    m2 : array_like
        Second matrix to be combined, should be of shape (2, 3).

    Returns
    -------
    np.ndarray
        The resulting matrix of shape (2, 3) after matrix
        multiplication.

    """
    # extend to a 3x3 for matrix multiplication
    _m1 = np.append(m1, np.array([[0.0, 0.0, 1.0]]), axis=0)
    _m2 = np.append(m2, np.array([[0.0, 0.0, 1.0]]), axis=0)
    m_combi = _m1.dot(_m2)[0:2]
    return m_combi


def _get_aoi_corners(dst_corners, resolution=None):
    polygon = Polygon(dst_corners)
    coords = np.array(polygon.exterior.coords)
    # estimate the angle of the bounding box
    # retrieve average line across AOI
    point1 = (coords[0] + coords[3]) / 2
    point2 = (coords[1] + coords[2]) / 2
    diff = point2 - point1
    angle = np.arctan2(diff[1], diff[0])
    # rotate the polygon over this angle to get a proper bounding box
    polygon_rotate = rotate(polygon, -angle, origin=tuple(dst_corners[0]), use_radians=True)

    xmin, ymin, xmax, ymax = polygon_rotate.bounds
    if resolution is not None:
        xmin = helpers.round_to_multiple(xmin, resolution)
        xmax = helpers.round_to_multiple(xmax, resolution)
        ymin = helpers.round_to_multiple(ymin, resolution)
        ymax = helpers.round_to_multiple(ymax, resolution)

    bbox_coords = [(xmin, ymax), (xmax, ymax), (xmax, ymin), (xmin, ymin), (xmin, ymax)]
    bbox = Polygon(bbox_coords)
    # now rotate back
    bbox = rotate(bbox, angle, origin=tuple(dst_corners[0]), use_radians=True)
    return bbox


def _get_aoi_width_length(dst_corners):
    points = [Point(x, y) for x, y, _ in dst_corners]
    linecross = LineString([points[0], points[1]])
    # linecross = LineString(dst_corners[0:2])
    length = np.abs(_get_perpendicular_distance(points[-1], linecross))
    point1 = np.array(dst_corners[0][0:2])
    point2 = np.array(dst_corners[1][0:2])
    diff = np.array(point2 - point1)
    angle = np.arctan2(diff[1], diff[0])

    # compute xy distance from line to other line making up the bounding box
    xy_diff = np.array([np.sin(-angle) * length, np.cos(angle) * length])
    points_pol = np.array([point1 - xy_diff, point1 + xy_diff, point2 + xy_diff, point2 - xy_diff])
    # always make sure the order of the points of upstream-left, downstream-left, downstream-right, upstream-right
    # if length <= 0:
    #     # negative length means the selected length is selected upstream of left-right cross section
    #     points_pol = np.array([point1 + xy_diff, point1, point2, point2 + xy_diff])
    # else:
    #     # postive means it is selected downstream of left-right cross section
    #     points_pol = np.array([point1, point1 + xy_diff, point2 + xy_diff, point2])

    return Polygon(points_pol)


def _smooth(img, stride):
    """Blur image through gaussian smoothing.

    Parameters
    ----------
    img: np.ndarray
        Input image.
    stride: int
        Size of the kernel used for the Gaussian blur.

    Returns
    -------
    np.ndarray
        blurred image.

    """
    blur = cv2.GaussianBlur(img.astype("float32"), (stride, stride), 0)
    return blur


def _convert_edge(img, stride_1, stride_2):
    """Enhance image gradients by a band filter.

    Parameters
    ----------
    img : np.ndarray
        Input image.
    stride_1 : int
        Size of the kernel used for the first Gaussian blur.
    stride_2 : int
        Size of the kernel used for the second Gaussian blur.

    Returns
    -------
    np.ndarray
        Edge enhanced image.

    """
    blur1 = cv2.GaussianBlur(img.astype("float32"), (stride_1, stride_1), 0)
    blur2 = cv2.GaussianBlur(img.astype("float32"), (stride_2, stride_2), 0)
    edges = blur2 - blur1
    return edges


def _get_dist_coefs(k1):
    """Convert barrel distortion parameter into distortion coefficient matrix.

    Parameters
    ----------
    k1 : float
        barrel lens distortion parameter

    Returns
    -------
    np.ndarray
        distortion coefficient matrix (4 parameter)

    """
    # define distortion coefficient vector
    dist = np.zeros((4, 1), np.float64)
    dist[0, 0] = k1
    return dist


def _get_perpendicular_distance(point, line):
    """Calculate perpendicular distance from point to line.

    Line is extended if perpendicular distance is larger than the distance to the endpoints.

    Parameters
    ----------
    point : shapely.geometry.Point
        x, y coordinates of point
    line : shapely.geometry.LineString
        line to calculate distance to

    Returns
    -------
    float
        perpendicular distance from point to line

    """
    # Get coordinates of line endpoints
    p1 = np.array(line.coords[0])
    p2 = np.array(line.coords[1])
    # Convert point to numpy array
    p3 = np.array(point.coords[0])

    # Calculate line vector
    line_vector = p2 - p1
    # Calculate vector from point to line start
    point_vector = p3 - p1

    # Calculate unit vector of line
    unit_line = line_vector / np.linalg.norm(line_vector)

    # Calculate projection length
    projection_length = np.dot(point_vector, unit_line)

    # Calculate perpendicular vector
    perpendicular_vector = point_vector - projection_length * unit_line
    perpendicular_distance = np.linalg.norm(perpendicular_vector)

    # Use cross product to calculate side
    cross_product = np.cross(line_vector, point_vector)

    # Determine the sign of the perpendicular distance
    return perpendicular_distance if cross_product > 0 else -perpendicular_distance


def get_cam_mtx(height, width, c=2.0, focal_length=None):
    """Compute camera matrix based on the given parameters for height, width, scaling factor, and focal length.

    Parameters
    ----------
    height : int
        The height of the image or frame from which the camera matrix is derived.
    width : int
        The width of the image or frame from which the camera matrix is derived.
    c : float, optional
        The scaling factor used to determine the principal point in the matrix.
        Default value is 2.0.
    focal_length : float, optional
        The focal length to set in the camera matrix. If not provided, defaults
        to the width of the image for both the x and y directions.

    Returns
    -------
    numpy.ndarray
        A 3x3 camera matrix defining the intrinsic parameters of the camera,
        including the principal point and focal length.

    """
    # define camera matrix
    mtx = np.eye(3, dtype=np.float32)
    mtx[0, 2] = width / c  # define center x
    mtx[1, 2] = height / c  # define center y
    if focal_length is None:
        mtx[0, 0] = width  # define focal length x
        mtx[1, 1] = width  # define focal length y
    else:
        mtx[0, 0] = focal_length  # define focal length x
        mtx[1, 1] = focal_length  # define focal length y
    return mtx


def get_ms_gftt(cap, start_frame=0, end_frame=None, n_pts=None, split=2, mask=None, wdw=4, progress=True):
    """Calculate motion smoothing of video frames using Good Features to Track and Lucas-Kanade Optical Flow methods.

    This function processes each frame between `start_frame` and `end_frame` to estimate and smooth affine
    transformations, which indicate motions. The function supports frame splitting for detecting features,
    applying an optional mask, and defines a smoothing window for the transformations.

    Parameters
    ----------
    cap : cv2.VideoCapture
        Video capture object to read frames from.
    start_frame : int, optional
        Index of the starting frame for processing. Defaults to 0.
    end_frame : int, optional
        Index of the ending frame for processing. If not specified, it is set to the last frame.
    n_pts : int, optional
        Number of good features to track within each frame. If None, the square root of
        the number of pixels in a frame is used.
    split : int, optional
        Number of segments to split each frame into for feature detection. Defaults to 2. Applied over both x and y
        direction.
    mask : np.ndarray, optional
        Optional mask to specify regions of interest within the frame for feature detection.
    wdw : int, optional
        Window size for smoothing the affine transformations over time. Defaults to 4.
    progress : bool, optional
        Show progress bar or not. Defaults to True.

    Returns
    -------
    list of ndarray
        A list of affine transformation matrices for each processed frame after smoothing.

    """
    # set end_frame to last if not defined
    end_frame = int(cap.get(cv2.CAP_PROP_FRAME_COUNT)) if end_frame is None else end_frame
    # make a start transform which does not change the first frame
    m = np.eye(3)[0:2]
    ms = []
    m_key = copy.deepcopy(m)
    # get start frame and points
    cap.set(cv2.CAP_PROP_POS_FRAMES, start_frame)
    # determine the mount of frames that must be processed
    n_frames = int(end_frame + 1) - int(start_frame)

    # Read first frame
    _, img_key = cap.read()
    # Convert frame to grayscale
    img1 = cv2.cvtColor(img_key, cv2.COLOR_BGR2GRAY)
    img_key = img1

    if n_pts is None:
        # use the square root of nr of pixels in a frame to decide on n_pts
        n_pts = int(np.sqrt(len(img_key.flatten())))

    # get features from first key frame
    prev_pts = _gftt_split(img_key, split, n_pts, mask=mask)

    pbar = tqdm(range(n_frames - 1), position=0, leave=True, disable=not (progress))
    pbar.set_description("Deriving stabilization parameters from second frame onwards")
    for i in pbar:
        ms.append(m)
        _, img2 = cap.read()
        img2 = cv2.cvtColor(img2, cv2.COLOR_BGR2GRAY)
        curr_pts, status, err = cv2.calcOpticalFlowPyrLK(img_key, img2, prev_pts, None)
        m_part = cv2.estimateAffine2D(curr_pts, prev_pts)[0]
        m = _combine_m(m_key, m_part)
        if i % 30 == 0:
            img_key = img1
            prev_pts = _gftt_split(img_key, split, n_pts, mask=mask)
            m_key = copy.deepcopy(m)
        img1 = img2
    # add the very last transformation
    ms.append(m)
    # smooth the affines over time
    ma = np.array(ms)
    for m in range(ma.shape[1]):
        for n in range(ma.shape[2]):
            ma[wdw:-wdw, m, n] = np.convolve(ma[:, m, n], np.ones(wdw * 2 + 1) / (wdw * 2 + 1), mode="valid")
    ms_smooth = list(ma)
    return ms_smooth


def _get_gcps_2_4(src, dst, img_width, img_height):
    """Convert 2 points GCPs to 4 points GCPs in the corners of the image frame.

    Function is used in case a nadir video is provided, where it may be assumed that transformation can be done
    through scaling and translation.

    Parameters
    ----------
    src : list or array-like
        source control points (list of lists)
    dst : list or array-like
        destination control points (list of lists)
    img_width : int
        width of original image frame
    img_height : int
        height of original image frame

    Returns
    -------
    src : list or array-like
        source control points (list of lists) converted to corner points
    dst : list or array-like
        destination control points (list of lists) converted to corner points

    """
    # first reverse the coordinate order of the y-axis
    _src = [[x, img_height - y] for x, y in src]
    # estimate affine transform (only rotation and translation)
    M = cv2.estimateAffinePartial2D(np.array(_src), np.array(dst))
    # complete M with a line indicating no perspective
    M = np.array(M[0].tolist() + [[0, 0, 1]])
    # establish corner coordinates
    corners = [[0, 0], [img_width, 0], [img_width, img_height], [0, img_height]]
    dst = cv2.perspectiveTransform(np.float32([corners]), M)[0].tolist()
    # now get the corners back transformed to the real image coordinates
    src = [[x, img_height - y] for x, y in corners]
    return src, dst


def _get_shape(bbox, resolution=0.01, round=1):
    """Calculate the number of columns and rows based on the dimensions of bounding box and provided resolution.

    Rounding factor can be used to adjust the number of columns and rows to a specified granularity.

    Parameters
    ----------
    bbox : shapely.geometry.Polygon
        Rectangular bounding box of AOI.
    resolution : float, optional
        x and y resolution in meters used to define grid
    round : int, optional
        A rounding factor that allows adjusting the number of columns and rows to be multiples of this factor.

    Returns
    -------
    tuple of int
        A tuple containing the number of columns and rows (cols, rows)
        calculated based on the bounding box dimensions, resolution, and
        rounding factor.

    """
    coords = bbox.exterior.coords
    box_length = LineString(coords[0:2]).length
    box_width = LineString(coords[1:3]).length
    cols = int(np.round((box_length / resolution) / round)) * round
    rows = int(np.round((box_width / resolution) / round)) * round
    return cols, rows


def _get_transform(bbox, resolution=0.01):
    """Return a rotated Affine transformation that fits with the bounding box and resolution.

    Parameters
    ----------
    bbox : shapely.geometry.Polygon
        polygon of bounding box. The coordinate order is very important and has to be:
        (upstream-left, downstream-left, downstream-right, upstream-right, upstream-left)
    resolution : float, optional
        resolution of target grid in meters (default: 0.01)

    Returns
    -------
    affine : rasterio.transform.Affine

    """
    corners = np.array(bbox.exterior.coords)
    # estimate the angle of the bounding box
    top_left_x, top_left_y = corners[0]
    # retrieve average line across AOI
    point1 = corners[0]
    point2 = corners[1]
    diff = point2 - point1
    # compute the angle of the projected bbox area of interest
    angle = np.arctan2(diff[1], diff[0])
    # compute per col the x and y diff
    dx_col, dy_col = np.cos(angle) * resolution, np.sin(angle) * resolution
    # compute per row the x and y offsets
    dx_row, dy_row = (
        np.cos(angle + 1.5 * np.pi) * resolution,
        np.sin(angle + 1.5 * np.pi) * resolution,
    )
    return rasterio.transform.Affine(dx_col, dy_col, top_left_x, dx_row, dy_row, top_left_y)


def _gftt_split(img, split, n_pts, mask=None):
    # split image in smaller chunks if user wants
    v = 0
    h = 0
    ver_split, hor_split = np.int16(np.ceil(np.array(img.shape) / split))
    pts = np.zeros((0, 1, 2), np.float32)
    while v < img.shape[0]:
        while h < img.shape[1]:
            sub_img = img[v : v + ver_split, h : h + hor_split]
            # get points over several quadrants
            subimg_pts = cv2.goodFeaturesToTrack(
                sub_img,
                mask=mask[v : v + ver_split, h : h + hor_split] if mask is not None else None,
                maxCorners=int(n_pts / split**2),
                qualityLevel=0.3,
                minDistance=10,
                blockSize=1,
            )
            # add offsets for quadrants
            if subimg_pts is not None:
                subimg_pts[:, :, 0] += h
                subimg_pts[:, :, 1] += v
                pts = np.append(pts, subimg_pts, axis=0)
            h += hor_split
        h = 0
        v += ver_split
    return pts


def solvepnp(dst, src, camera_matrix, dist_coeffs):
    """Solve p-n-p problem.

    Wrapper for cv2.SolvePnP with pre-processing of the input data and selection of the correct flags.

    Parameters
    ----------
    src : list of lists
        [x, y] with source coordinates, typically cols and rows in image
    dst : list of lists
        [x, y] (in case of 4 planar points) or [x, y, z] (in case of 6+ 3D points) with target coordinates after
        reprojection, can e.g. be in crs [m]
    camera_matrix : np.ndarray (3x3)
        Camera intrinsic matrix
    dist_coeffs : p.ndarray, optional
        1xN array with distortion coefficients (N = 4, 5 or 8)

    Returns
    -------
    succes : int
        0, 1 for succes or no succes
    rvec : np.array
        rotation vector
    tvec : np.array
        translation vector

    """
    # set points to float32
    _src = np.float64(src)
    _dst = np.float64(dst)

    if len(_dst) == 4:
        flags = cv2.SOLVEPNP_P3P
        # if 4 x, y points are provided, add a column with zeros
        # _dst = np.c_[_dst, np.zeros(4)]
    else:
        flags = cv2.SOLVEPNP_ITERATIVE

    camera_matrix = np.float32(camera_matrix)
    dist_coeffs = np.float32(dist_coeffs)
    # define transformation matrix based on GCPs
    return cv2.solvePnP(_dst, _src, camera_matrix, dist_coeffs, flags=flags)


def transform(img, m):
    """Affine-transform image using specified affine transform matrix.

    Typically the transformation is derived for image stabilization purposes.

    Parameters
    ----------
    img : np.ndarray 2D [MxN] or 3D [MxNx3] if RGB image
        Image used as input
    m : np.ndarray [2x3]
        Affine transformation

    Returns
    -------
    img_transform : np.ndarray 2D [MxN] or 3D [MxNx3] if RGB image
        Image after affine transform applied

    """
    h = img.shape[0]
    w = img.shape[1]
    # Apply affine wrapping to the given frame
    img_transform = cv2.warpAffine(img, m, (w, h))
    return img_transform


def calibrate_camera(
    fn,
    chessboard_size=(9, 6),
    max_imgs=30,
    plot=True,
    progress_bar=True,
    criteria=criteria,
    to_file=False,
    frame_limit=None,
    tolerance=None,
):
    """Calculate intrinsic matrix and distortion coefficients.

    This follows recipe from:
    https://docs.opencv.org/4.x/dc/dbb/tutorial_py_calibration.html
    """
    dir = os.path.split(os.path.abspath(fn))[0]
    cap = cv2.VideoCapture(fn)
    # make a list of logical frames in order to read
    frames_count = cap.get(cv2.CAP_PROP_FRAME_COUNT)
    frames_list = helpers.staggered_index(start=0, end=frames_count - 1)

    # set the expected object points from the chessboard size
    objp = np.zeros((chessboard_size[0] * chessboard_size[1], 3), np.float32)
    objp[:, :2] = np.mgrid[0 : chessboard_size[0], 0 : chessboard_size[1]].T.reshape(-1, 2)

    obj_pts = []
    img_pts = []
    imgs = []

    ret_img, img = cap.read()
    frame_size = img.shape[1], img.shape[0]
    if frame_limit is not None:
        frames_list = frames_list[0:frame_limit]
    if progress_bar:
        frames_list = tqdm(frames_list, position=0, leave=True)
        # pbar.update(1)
    for f in frames_list:
        cap.set(cv2.CAP_PROP_POS_FRAMES, f)
        ret_img, img = cap.read()
        if ret_img:
            gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
            # cur_f = cap.get(cv2.CAP_PROP_POS_FRAMES)
            ret, corners = cv2.findChessboardCorners(gray, chessboard_size, flags=cv2.CALIB_CB_FAST_CHECK)
            if ret:
                # append the expected point coordinates
                obj_pts.append(objp)
                imgs.append(copy.deepcopy(img))
                corners2 = cv2.cornerSubPix(gray, corners, (11, 11), (-1, -1), criteria)
                img_pts.append(corners)
                cv2.drawChessboardCorners(img, chessboard_size, corners2, ret)
                # add frame number
                cv2.putText(img, f"Frame {f}", (20, 100), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 0, 0), 8, 2)
                cv2.putText(img, f"Frame {f}", (20, 100), cv2.FONT_HERSHEY_SIMPLEX, 1, (255, 255, 255), 2, 2)
                height = 720
                width = int(img.shape[1] * height / img.shape[0])
                imS = cv2.resize(img, (width, height), interpolation=cv2.INTER_AREA)
                if plot:
                    cv2.imshow("img", imS)
                    cv2.waitKey(500)
                if to_file:
                    # write a small version to jpg
                    jpg = os.path.join(dir, "frame_{:06d}.png".format(int(f)))
                    cv2.imwrite(jpg, imS)

                if len(imgs) == max_imgs:
                    print(f"Maximum required images {max_imgs} found")
                    break
    if progress_bar:
        frames_list.close()

    cap.release()
    # close the plot window if relevant
    cv2.destroyAllWindows()
    # do calibration
    assert len(obj_pts) >= 5, (
        f"A minimum of 5 frames with chessboard patterns must be available, only {len(obj_pts)} found. Please check "
        f"if the video contains chessboard patterns of size {chessboard_size} "
    )
    ret, camera_matrix, dist_coeffs, rvecs, tvecs = cv2.calibrateCamera(obj_pts, img_pts, frame_size, None, None)
    # remove badly performing images and recalibrate
    errs = []
    for i in range(len(obj_pts)):
        img_pts2, _ = cv2.projectPoints(obj_pts[i], rvecs[i], tvecs[i], camera_matrix, dist_coeffs)
        errs.append(cv2.norm(img_pts[i], img_pts2, cv2.NORM_L2) / len(img_pts2))

    if tolerance is not None:
        # remove high error
        idx = np.array(errs) < tolerance
        obj_pts = list(np.array(obj_pts)[idx])
        img_pts = list(np.array(img_pts)[idx])
        # do calibration
        ret, camera_matrix, dist_coeffs, rvecs, tvecs = cv2.calibrateCamera(obj_pts, img_pts, frame_size, None, None)
        errs = []
        for i in range(len(obj_pts)):
            img_pts2, _ = cv2.projectPoints(obj_pts[i], rvecs[i], tvecs[i], camera_matrix, dist_coeffs)
            errs.append(cv2.norm(img_pts[i], img_pts2, cv2.NORM_L2) / len(img_pts2))
    print(f"Average error on point reconstruction is {np.array(errs).mean()}")
    return camera_matrix, dist_coeffs


def _Rt_to_M(rvec, tvec, camera_matrix, z=0.0, reverse=False):
    R = cv2.Rodrigues(rvec)[0]
    # assume height of projection plane
    R[:, 2] = R[:, 2] * z
    # add translation vector
    R[:, 2] = R[:, 2] + tvec.flatten()
    # compute homography
    if reverse:
        # From perspective to objective
        M = np.dot(camera_matrix, R)
    else:
        # from objective to perspective
        M = np.linalg.inv(np.dot(camera_matrix, R))
    # normalize homography before returning
    return M / M[-1, -1]
    return M / M[-1, -1]


def pose_world_to_camera(rvec, tvec):
    """Convert a world coordinate pose to a camera coordinate pose or vice versa.

    Parameters
    ----------
    rvec : numpy.ndarray
        A 3x1 or 1x3 rotation vector in world coordinates.
    tvec : numpy.ndarray
        A 3x1 translation vector in world coordinates.

    Returns
    -------
    tuple
        A tuple containing:
        - rvec (numpy.ndarray): A 1-dimensional array representing the rotation vector
          in camera coordinates.
        - tvec (numpy.ndarray): A 3x1 array representing the translation vector
          in camera coordinates.

    """
    # Get Rodriguez rotation matrix
    R_input, _ = cv2.Rodrigues(rvec.flatten())

    # 2. Create the OpenCV-compatible transformation
    R = R_input.T  # Transpose of the rotation matrix
    tvec = -R @ tvec.flatten()  # Transform the translation vector

    # 3. Convert the adjusted rotation matrix back to a rotation vector
    rvec, _ = cv2.Rodrigues(R)
    rvec = rvec.flatten()
    return rvec, tvec


def distort_points(points, camera_matrix, dist_coeffs):
    """Distort x, y point locations with provided lens parameters.

    Points can be back projected on original (distorted) frame positions.

    Adapted from https://answers.opencv.org/question/148670/re-distorting-a-set-of-points-after-camera-calibration/

    Parameters
    ----------
    points : list of lists
        undistorted points [x, y], provided as float
    camera_matrix : array-like [3x3]
        camera matrix
    dist_coeffs : array-like [4]
        distortion coefficients

    Returns
    -------
    points : list of lists
        distorted point coordinates [x, y] as floats

    """
    points = np.array(points, dtype=np.float64)
    # ptsTemp = np.array([], dtype='float32')
    # make empty rotation and translation vectors (we are only undistorting)
    rtemp = ttemp = np.array([0, 0, 0], dtype="float32")
    # normalize the points to be independent of the camera matrix using undistortPoints with no distortion matrix
    ptsOut = cv2.undistortPoints(points, camera_matrix, None)
    # convert points to 3d points
    ptsTemp = cv2.convertPointsToHomogeneous(ptsOut)
    # project them back to image space using the distortion matrix
    return np.int32(
        np.round(cv2.projectPoints(ptsTemp, rtemp, ttemp, camera_matrix, dist_coeffs, ptsOut)[0][:, 0])
    ).tolist()


def get_M_2D(src, dst, reverse=False):
    """Retrieve homography matrix for between (4) src and (4) dst points with only x, y coordinates (no z).

    Parameters
    ----------
    src : list of lists
        [x, y] with source coordinates, typically cols and rows in image
    dst : list of lists
        [x, y] with target coordinates after reprojection, can e.g. be in crs [m]
    reverse : bool, optional
        If set, the reverse homography to back-project to camera objective will be retrieved

    Returns
    -------
    M : np.ndarray
        homography matrix (3x3), used in cv2.warpPerspective

    """
    # set points to float32
    _src = np.float32(src)
    _dst = np.float32(dst)
    # define transformation matrix based on GCPs
    if reverse:
        M = cv2.getPerspectiveTransform(_dst, _src)
    else:
        M = cv2.getPerspectiveTransform(_src, _dst)
    return M


def get_M_3D(src, dst, camera_matrix, dist_coeffs=None, z=0.0, reverse=False):
    """Retrieve homography matrix for between (6+) 2D src and (6+) 3D dst (x, y, z) points.

    Parameters
    ----------
    src : list of lists
        [x, y] with source coordinates, typically cols and rows in image
    dst : list of lists
        [x, y, z] with target coordinates after reprojection, can e.g. be in crs [m]
    camera_matrix : np.ndarray (3x3)
        Camera intrinsic matrix
    dist_coeffs : p.ndarray, optional
        1xN array with distortion coefficients (N = 4, 5 or 8)
    z : float, optional
        Elevation plane (real-world coordinate crs) of projected image
    reverse : bool, optional
        If set, the reverse homography to back-project to camera objective will be retrieved

    Returns
    -------
    M : np.ndarray
        homography matrix (3x3), used in cv2.warpPerspective

    Notes
    -----
    See rectification workflow OpenCV
    http://docs.opencv.org/modules/calib3d/doc/camera_calibration_and_3d_reconstruction.html
    Code based on:
    https://www.openearth.nl/flamingo/_modules/flamingo/rectification/rectification.html

    """
    dist_coeffs = np.zeros((1, 4)) if dist_coeffs is None else dist_coeffs
    success, rvec, tvec = solvepnp(dst, src, camera_matrix, dist_coeffs)
    return _Rt_to_M(rvec, tvec, camera_matrix, z=z, reverse=reverse)


def color_scale(img, method):
    """Transform color space of an image according to the specified method.

    The function offers support for grayscale conversion, RGB transformation
    for plotting, and HSV conversion. Images initially are expected to be in
    BGR format and will remain unchanged if 'bgr' is specified as the method.

    Parameters
    ----------
    img : np.ndarray
        Input image in BGR color space [uint8 type).
    method : str
        The color space conversion to apply to the image. Supported methods are
        'grayscale', 'rgb', 'hsv', and 'bgr'.

    Returns
    -------
    np.ndarray
        Image transformed into the specified color space.

    """
    if method == "grayscale":
        # apply gray scaling, contrast- and gamma correction
        # img = _corr_color(img, alpha=None, beta=None, gamma=0.4)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)  # mean(axis=2)
    elif method == "rgb":
        # turn bgr to rgb for plotting purposes
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    elif method == "hsv":
        img = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
    elif method == "bgr":
        pass
    elif method == "hue":
        img = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)[:, :, 0]
    elif method == "sat":
        img = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)[:, :, 1]
    elif method == "val":
        img = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)[:, :, 2]

    return img


def get_frame(cap, rotation=None, ms=None, method="grayscale"):
    """Get single frame from video capture.

    The function captures an image frame from the given video capture device
    and applies optional rotation, stabilization, and color scaling.

    Parameters
    ----------
    cap : VideoCapture
        The video capture object from which the frame is to be read.
    rotation : int, optional
        Specifies the angle for the rotation of the image. If None, no
        rotation is applied.
    ms : np.ndarray, optional
        2x3 affina transformation parameters. If None, the image will not undergo stabilization.
    method : str, optional
        Describes the color scaling method to be applied on the image.
        Default is "grayscale". Can also be "rgb", "hsv", "bgr", "hue", "sat", or "val".

    Returns
    -------
    ret : bool
        A flag indicating whether a frame was successfully read.
    img : ndarray
        The image after applying the optional processing steps.

    Raises
    ------
    IOError
        If the function is unable to read from the capture device, an
        IOError is raised indicating a possible issue with the stream.

    """
    try:
        ret, img = cap.read()
        if rotation is not None:
            img = cv2.rotate(img, rotation)
    except IOError:
        raise IOError("Cannot read")
    if ret:
        if ms is not None:
            # apply stabilization on image
            img = transform(img, ms)
        img = color_scale(img, method)
    return ret, img


def get_time_frames(cap, start_frame, end_frame, lazy=True, fps=None, progress=True, **kwargs):
    """Obtain valid time stamps and frame numbers from video capture object.

    Valid frames may start and end at start_frame and end_frame, respectively. However, certain required frames may
    turn out not readable. These will be captured, and resulting set of valid frames returned only.

    Parameters
    ----------
    cap : cv2.VideoCapture
        Opened VideoCapture object
    start_frame : int
        first frame to consider for reading
    end_frame : int
        last frame to consider for reading
    lazy : bool, optional
        read frames lazily (default) or not. Set to False for direct reading (faster, but more memory)
    fps : float, optional
        hard enforced frames per second number (used when metadata of video is incorrect)
    progress : bool, optional
        display progress bar. Default is True.
    **kwargs : dict, optional
        additional keyword arguments passed to get_frame() function

    Returns
    -------
    time : list
        list with valid time stamps in milliseconds. time stamps belongs to the start of the frame in frame number.
    frame_number : list
        list with valid frame numbers

    """
    cap.set(cv2.CAP_PROP_POS_FRAMES, np.float64(start_frame))
    pbar = tqdm(
        total=end_frame - start_frame + 1, position=0, desc="Scanning video", disable=not (progress), leave=True
    )
    ret, img = get_frame(cap, **kwargs)
    # pbar.update(1)
    n = start_frame
    time = []
    frame_number = []
    frames = None if lazy else []
    while ret:
        if n > end_frame:
            break
        if not lazy:
            frames.append(img)
        t1 = cap.get(cv2.CAP_PROP_POS_MSEC)
        time.append(n * 1000.0 / fps) if fps is not None else time.append(t1)
        frame_number.append(n)
        n += 1
        ret, img = get_frame(cap, **kwargs)  # read frame 1 + ...
        pbar.update(1)
        if ret == False:
            break
        t2 = cap.get(cv2.CAP_PROP_POS_MSEC)
        if t2 <= 0.0:
            # invalid time difference, stop reading.
            break
    # do a final check if the last frame(s) are readable by direct seek and read. Sometimes this results in not being
    # able to r
    last_valid_idx = _check_valid_frames(cap, frame_number)
    if last_valid_idx is not None:
        time = time[:last_valid_idx]
        frame_number = frame_number[:last_valid_idx]
        if not lazy:
            frames = frames[:last_valid_idx]
    return time, frame_number, frames


def get_ortho(img, M, shape, flags=cv2.INTER_AREA):
    """Reproject an image to a given shape using perspective transformation matrix M.

    Parameters
    ----------
    img: np.ndarray
        image to transform
    M: np.ndarray
        image perspective transformation matrix
    shape: tuple of ints
        (cols, rows)
    flags: cv2.flags
        passed with cv2.warpPerspective

    Returns
    -------
        img : np.ndarray
            reprojected data with shape=shape

    """
    return cv2.warpPerspective(img, M, shape, flags=flags)


def get_aoi(dst_corners, resolution=None, method="corners"):
    """Get rectangular AOI from 3 or 4 user defined points within frames.

    Parameters
    ----------
    dst_corners : np.ndarray
        corners of aoi, with `method="width_length"` in order: left-bank, right-bank, up/downstream point,
        with `method="corners"` in order: upstream-left, downstream-left, downstream-right, upstream-right.
    resolution : float
        resolution of intended reprojection, used to round the bbox to a whole number of intended pixels
    method : str
        can be "corners" or "width_length". With "corners", the AOI is defined by the four corners of the rectangle.
        With "width" length, the AOI is defined by the width (2 points) and length (1 point) of the rectangle.

    Returns
    -------
    bbox : shapely.geometry.Polygon
        bounding box of aoi (with rotated affine)

    """
    if method == "corners":
        bbox = _get_aoi_corners(dst_corners, resolution)
    elif method == "width_length":
        bbox = _get_aoi_width_length(dst_corners)

    else:
        raise ValueError("method must be 'corners' or 'width_length'")

    return bbox


def get_polygon_pixels(img, pol, reverse_y=False):
    """Get pixel intensities within a polygon."""
    if pol.is_empty:
        return np.array([np.nan])
    polygon_coords = list(pol.exterior.coords)
    mask = np.zeros_like(img, dtype=np.uint8)
    cv2.fillPoly(mask, np.array([polygon_coords], np.int32), color=255)
    if reverse_y:
        return np.flipud(img)[mask == 255]
    return img[mask == 255]


def optimize_intrinsic(src, dst, height, width, c=2.0, lens_position=None, camera_matrix=None, dist_coeffs=None):
    """Optimize the intrinsic parameters of a camera model.

    The function finds optimal intrinsic camera parameters, including focal length and distortion coefficients, by
    minimizing the reprojection error from 3D source points to 2D destination points. It uses differential evolution
    for optimization. Optionally lens position can be provided to include additional geometric constraints.

    Parameters
    ----------
    src : array_like
        Source points in the original 3D space to be projected.
    dst : array_like
        Destination points in the 2D image space, serving as the target for
        projecting the source points.
    height : int
        The height of the image in pixels.
    width : int
        The width of the image in pixels.
    c : float, optional
        Center parameter of the camera matrix.
    lens_position : array_like, optional
        The assumed position of the lens in the 3D space.
    camera_matrix : Optional[List[List]]
        Predefined camera matrix. If not provided focal length will be fitted and camera matrix returned
    dist_coeffs : Optional[List[List]]
        Distortion coefficients to be used for the camera. If not provided, the first two (k1, k2)
        distortion coefficients are fitted on data.

    Returns
    -------
    Tuple[numpy.ndarray, numpy.ndarray, float]
        A tuple containing the optimized camera matrix, distortion coefficients,
        and the minimized reprojection error.

    """

    def error_intrinsic(x, src, dst, height, width, c=2.0, lens_position=None, camera_matrix=None, dist_coeffs=None):
        """Compute the reprojection error for the intrinsic parameters of a camera model.

        This function optimizes for the focal length and first two distortion coefficients based on the source and
        destination point correspondences provided. Lens position may be provided as additional geometric constraint.

        Parameters
        ----------
        x : array_like
            The array containing the optimization parameters, where `x[0]` is used
            to compute the focal length and `x[1]` and `x[2]` are used to adjust
            the distortion coefficients.
        src : array_like
            Source points in the original 3D space that need to be projected.
        dst : array_like
            Destination points in the 2D image space, which are the target for
            the projection of the source points.
        height : int
            The height of the image in pixels.
        width : int
            The width of the image in pixels.
        c : float, optional
            center parameter of camera matrix.
        lens_position : array_like, optional
            The assumed position of the lens in the 3D space.
        camera_matrix : array_like, optional
            camera matrix [3x3]
        dist_coeffs : array_like, optional
            Distortion coefficients.

        Returns
        -------
        float
            The computed mean reprojection error, with optional contributions from
            the camera position error if the lens position is provided.

        """
        param_nr = 0
        # set the parameters
        if camera_matrix is None:
            f = x[param_nr] * width
            camera_matrix_sample = get_cam_mtx(height, width, c=c, focal_length=f)
            param_nr += 1
        else:
            # take the existing camera matrix
            camera_matrix_sample = camera_matrix.copy()
        if dist_coeffs is None:
            dist_coeffs_sample = DIST_COEFFS.copy()
            k1 = x[param_nr]
            k2 = x[param_nr + 1]
            dist_coeffs_sample[0][0] = k1
            dist_coeffs_sample[1][0] = k2
        else:
            # take the existing distortion coefficients
            dist_coeffs_sample = dist_coeffs.copy()
            k1 = dist_coeffs_sample[0][0]
            k2 = dist_coeffs_sample[1][0]

        # initialize error
        err = 100
        cam_err = None

        # reduce problem space to centered around gcp average
        coord_mean = np.array(dst).mean(axis=0)
        _dst = np.float64(np.array(dst) - coord_mean)
        zs = np.zeros(4) if len(_dst[0]) == 2 else np.array(_dst)[:, -1]
        if lens_position is not None:
            _lens_pos = np.array(lens_position) - coord_mean

        # camera_matrix = _get_cam_mtx(height, width, c=c, focal_length=f)
        success, rvec, tvec = solvepnp(_dst, src, camera_matrix_sample, dist_coeffs_sample)
        if success:
            # estimate destination locations from pose
            dst_est = unproject_points(src, zs, rvec, tvec, camera_matrix_sample, dist_coeffs_sample)
            # src_est = np.array([list(point[0]) for point in src_est])
            dist_xy = np.array(_dst)[:, 0:2] - np.array(dst_est)[:, 0:2]
            dist = (dist_xy**2).sum(axis=1) ** 0.5
            gcp_err = dist.mean()
            if lens_position is not None:
                rmat = cv2.Rodrigues(rvec)[0]
                lens_pos2 = np.array(-rmat).T @ tvec
                cam_err = ((_lens_pos - lens_pos2.flatten()) ** 2).sum() ** 0.5
                # TODO: for now camera errors are weighted with 10% needs further investigation
            err = float(0.1 * cam_err + gcp_err) if cam_err is not None else gcp_err
        return err  # assuming gcp pixel distance is about 5 cm

    # determine optimization bounds
    bounds = []
    if camera_matrix is not None and dist_coeffs is not None:
        # both are already known, so nothing to do
        return camera_matrix, dist_coeffs, None
    if camera_matrix is None:
        bounds.append([float(0.25), float(2)])
    if len(dst) > 4 and dist_coeffs is None:
        bounds.append([-0.9, 0.9])  # k1
        bounds.append([-0.5, 0.5])  # k2
    else:
        # set a warning if dist_coeffs is provided without sufficient ground control
        if dist_coeffs:
            warnings.warn(
                "You are trying to optimize distortion coefficients with only 4 GCPs. "
                "This would lead to overfitting, setting distortion coefficients to zero.",
                stacklevel=2,
            )
        dist_coeffs = DIST_COEFFS.copy()
    # if len(dst) == 4:
    #     bnds_k1 = (-0.0, 0.0)
    #     bnds_k2 = (-0.0, 0.0)
    # else:
    #     # bnds_k1 = (-0.2501, -0.25)
    #     bnds_k1 = (-0.9, 0.9)
    #     bnds_k2 = (-0.5, 0.5)
    # bnds_k1 = (-0.0, 0.0)
    # bnds_k2 = (-0.0, 0.0)
    opt = optimize.differential_evolution(
        error_intrinsic,
        # bounds=[(float(0.25), float(2)), bnds_k1],#, (-0.5, 0.5)],
        bounds=bounds,
        # bounds=[(1710./width, 1714./width), bnds_k1, bnds_k2],
        args=(src, dst, height, width, c, lens_position, camera_matrix, dist_coeffs),
        atol=0.001,  # one mm
    )
    param_nr = 0
    if camera_matrix is None:
        camera_matrix = get_cam_mtx(height, width, focal_length=opt.x[param_nr] * width)
        # move to next parameter
        param_nr += 1
    if dist_coeffs is None:
        dist_coeffs = DIST_COEFFS
        dist_coeffs[0][0] = opt.x[param_nr]
        dist_coeffs[1][0] = opt.x[param_nr + 1]
    # dist_coeffs[4][0] = opt.x[3]
    # dist_coeffs[3][0] = opt.x[4]
    # print(f"CAMERA MATRIX: {camera_matrix}")
    # print(f"DIST COEFFS: {dist_coeffs}")
    return camera_matrix, dist_coeffs, opt.fun


def transform_to_bbox(coords, bbox, resolution):
    """Transform coordinates defined in crs of bbox, into cv2 compatible pixel coordinates.

    Parameters
    ----------
    coords : list of lists
        [x, y, z] with coordinates
    bbox : shapely Polygon
        Bounding box. The coordinate order is very important and has to be upstream-left, downstream-left,
        downstream-right, upstream-right, upstream-left
    resolution : float
        resolution of target pixels within bbox

    Returns
    -------
    colrows : list
        tuples of columns and rows

    """
    # first assemble x and y coordinates
    transform = _get_transform(bbox, resolution)
    if len(coords[0]) == 3:
        xs, ys, zs = zip(*coords)
    else:
        xs, ys = zip(*coords)
    rows, cols = rasterio.transform.rowcol(transform, xs, ys, op=float)
    return list(zip(cols, rows)) if len(coords[0]) == 2 else list(zip(cols, rows, zs))


def undistort_img(img, camera_matrix, dist_coeffs):
    """Correct lens distortion of image based on lens characteristics.

    Function by Gerben Gerritsen / Sten Schurer, 2019.

    Parameters
    ----------
    img : np.ndarray
        3D array with image
    camera_matrix: np.ndarray
        Camera matrix
    dist_coeffs: np.ndarray
        distortion coefficients

    Returns
    -------
    img: np.ndarray
        undistorted img

    """
    # correct image for lens distortion
    return cv2.undistort(img, np.array(camera_matrix), np.array(dist_coeffs))


def unproject_points(src, z, rvec, tvec, camera_matrix, dist_coeffs):
    """Reverse-project points from the image to the 3D world.

    As points on the objective are a ray line in the real world, a x, y, z coordinate can only be reconstructed if
    the points have one known coordinate (z).

    Parameters
    ----------
    src : list (of lists)
        pixel coordinates
    z : float or list of floats
        z-level belonging to src points (if list, then the length should be equal to the number of src points)
    rvec : array-like
        rotation vector
    tvec : array-like
        translation vector
    camera_matrix : array-like
        camera matrix
    dist_coeffs : array_like
        distortion coefficients

    Returns
    -------
    np.ndarray
        unprojected points (x, y, z)

    """
    src = np.float64(np.atleast_1d(src))
    # first undistort points
    src = np.float64(np.array(undistort_points(src, camera_matrix, dist_coeffs)))
    rvec = np.array(rvec)
    tvec = np.array(tvec)
    if isinstance(z, (list, np.ndarray)):
        #         zs = np.atleast_1d(zs)
        z = np.float64(z)
        dst = []
        assert len(z) == len(
            src
        ), f"Amount of src points {len(src)} is not equal to amount of vertical levels z {len(z)}"
        for pt, _z in zip(src, z):
            M = _Rt_to_M(rvec, tvec, camera_matrix, z=_z, reverse=False)
            x, y = list(cv2.perspectiveTransform(pt[None, None, ...], M)[0][0])
            dst.append([x, y, _z])
    else:
        # z is only one value, so there is no change in M. We can do this faster in one go
        M = _Rt_to_M(rvec, tvec, camera_matrix, z=z, reverse=False)
        dst = cv2.perspectiveTransform(src[None, ...], M)[0]
        dst = np.hstack(
            (
                dst,
                np.ones((len(dst), 1)) * z,
            )
        )
    return dst


def undistort_points(points, camera_matrix, dist_coeffs, reverse=False):
    """Undistort x, y point locations with provided lens parameters.

    points can be undistorted together with images from that lens.

    Parameters
    ----------
    points : list of lists
        points [x, y], provided as float
    camera_matrix : array-like [3x3]
        camera matrix
    dist_coeffs : array-like [4]
        distortion coefficients
    reverse : bool, optional
        if set, the distortion will be undone, so that points can be back projected on original (distorted) frame
        positions.

    Returns
    -------
    points : list of lists
        undistorted point coordinates [x, y] as floats

    """
    camera_matrix = np.array(camera_matrix)
    dist_coeffs = np.array(dist_coeffs)
    if reverse:
        return distort_points(points, camera_matrix, dist_coeffs)

    points_undistort = cv2.undistortPoints(
        np.expand_dims(np.float64(points), axis=1), camera_matrix, dist_coeffs, P=camera_matrix
    )
    return points_undistort[:, 0].tolist()


def world_to_camera(points: np.ndarray, rvec: np.ndarray, tvec: np.ndarray):
    """Transform points from the world coordinate system to the camera coordinate system.

    Parameters
    ----------
    points : np.ndarray
        An array of 3D points in the world coordinate system.
    rvec : np.ndarray
        OpenCV compatible rotation vector.
    tvec : np.ndarray
        OpenCV comptatible translation vector.

    Returns
    -------
    np.ndarray
        array of 3D points in the camera coordinate system.

    """
    # get Rodriguez rotation matrix
    R, _ = cv2.Rodrigues(rvec)
    # rotate and translate points
    return (points @ R.T) + tvec.T



"""Camera configuration for pyorc."""

import copy
import json
import os
import warnings
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

import cv2
import matplotlib.pyplot as plt
import numpy as np
import shapely.geometry
from pyproj import CRS, Transformer
from pyproj.exceptions import CRSError
from rasterio.features import rasterize
from shapely import ops, wkt
from shapely.geometry import LineString, Point, Polygon

from pyorc import cv, helpers, plot_helpers

MODES = Literal["camera", "geographical", "3d"]


class CameraConfig:
    """Camera configuration containing information about the perspective of the camera.

    The camera configuration contains information and functionalities to reconstruct perspective information
    relating 2D image coordinates to 3D real world coordinates.
    """

    def __str__(self):
        return str(self.to_json())

    def __repr__(self):
        return self.to_json()

    def __init__(
        self,
        height: int,
        width: int,
        crs: Optional[Any] = None,
        window_size: int = 10,
        resolution: float = 0.05,
        bbox: Optional[Union[shapely.geometry.Polygon, str]] = None,
        camera_matrix: Optional[List[List[float]]] = None,
        dist_coeffs: Optional[List[List[float]]] = None,
        lens_position: Optional[List[float]] = None,
        corners: Optional[List[List[float]]] = None,
        gcps: Optional[Dict[str, Union[List, float]]] = None,
        calibration_video: Optional[str] = None,
        is_nadir: Optional[bool] = False,
        stabilize: Optional[List[List]] = None,
        rotation: Optional[int] = None,
        rvec: Optional[List[float]] = None,
        tvec: Optional[List[float]] = None,
    ):
        """Initialize a camera configuration instance.

        Parameters
        ----------
        height : int
            height of frame in pixels
        width : int
            width of frame in pixels
        crs : int, dict or str, optional
            Coordinate Reference System. Accepts EPSG codes (int or str) proj (str or dict) or wkt (str). Only used if
            the data has no native CRS.
        window_size : int
            pixel size of interrogation window (default: 15)
        resolution : float, optional
            resolution in m. of projected pixels (default: 0.01)
        bbox : shapely.geometry.Polygon, optional
            bounding box in geographical coordinates
        camera_matrix : List[List[float]], optional
            pre-defined camera matrix (if e.g. known from a separate calibration)
        dist_coeffs : List[List[float]], optional
            pre-defined distortion parameters (if e.g. known from a separate calibration)
        lens_position : list of floats (3),
            x, y, z coordinate of lens position in CRS
        corners : list of lists of floats (2)
            [x, y] coordinates defining corners of area of interest in camera cols/rows, bbox will be computed from this
        gcps : dict
            Can contain "src": list of lists, with column, row locations in objective of control points,
            "dst": list of lists, with x, y or x, y, z locations (local or global coordinate reference system) of
            control points,
            "h_ref": float, measured water level [m] in local reference system (e.g. from staff gauge or pressure gauge)
            during gcp survey,
            "z_0": float, water level [m] in global reference system (e.g. from used GPS system CRS). This must be in
            the same vertical reference as the measured bathymetry and other survey points,
            "crs": int, str or CRS object, CRS in which "dst" points are measured. If None, a local coordinate system is
            assumed (e.g. from spirit level).
        calibration_video : str, optional
            local path to video file containing a checkerboard pattern. Must be 9x6 if called directly, otherwise use
            ``.calibrate_camera`` explicitly and provide ``chessboard_size`` explicitly. When used, an automated camera
            calibration on the video file will be attempted.
        is_nadir : bool, optional
            If set, the video is assumed to be taken at sub-drone position and only two control points are needed
            for camera configuration
        stabilize : list (of lists), optional
            contains [x, y] pixels defining a polygon enclosing moving (water) areas. Areas outside of the polygon
            are used for stabilization of the video, if this polygon is defined.
        rotation : int [90, 180, 270]
            enforces a rotation of the video of 90, 180 or 2780 degrees clock-wise.
        rvec : list of floats (3), optional
            OpenCV compatible rotation vector, if known. If None, the rotation will be computed from pnp solving if
            gcps are available.
        tvec : list of floats (3), optional
            OpenCV compatible translation vector, if known. If None, the rotation will be computed from pnp solving if
            gcps are available.

        """
        assert isinstance(height, int), 'height must be provided as type "int"'
        assert isinstance(width, int), 'width must be provided as type "int"'
        assert isinstance(window_size, int), 'window_size must be of type "int"'
        self.height = height
        self.width = width
        self.is_nadir = is_nadir
        self.camera_matrix = camera_matrix
        self.dist_coeffs = dist_coeffs
        self.rvec = rvec
        self.tvec = tvec
        if crs is not None:
            try:
                crs = CRS.from_user_input(crs)
            except CRSError:
                raise CRSError(f'crs "{crs}" is not a valid Coordinate Reference System')
            assert crs.is_geographic == 0, "Prodstvided crs must be projected with units like [m]"
            self.crs = crs.to_wkt()
        if resolution is not None:
            self.resolution = resolution
        if lens_position is not None:
            self.set_lens_position(*lens_position)
        else:
            self.lens_position = None
        if gcps is not None:
            self.set_gcps(**gcps)
        if self.is_nadir:
            # with nadir, no perspective can be constructed, hence, camera matrix and dist coeffs will be set
            # to default values
            self.camera_matrix = cv.get_cam_mtx(self.height, self.width)
            self.dist_coeffs = cv.DIST_COEFFS
        # camera pars are incomplete and need to be derived
        else:
            self.calibrate()
        if calibration_video is not None:
            self.set_lens_calibration(calibration_video, plot=False)
        if bbox is not None:
            self.bbox = bbox
        if window_size is not None:
            self.window_size = window_size
        # override the transform and bbox with the set corners
        if corners is not None:
            self.set_bbox_from_corners(corners)
        if stabilize is not None:
            self.stabilize = stabilize
        if rotation is not None:
            self.rotation = rotation

    @property
    def bbox(self):
        """Give geographical bbox fitting around corners points of area of interest in camera perspective.

        Returns
        -------
        bbox : shapely.geometry.Polygon
            bbox of area of interest

        """
        return self._bbox

    @bbox.setter
    def bbox(self, pol):
        if isinstance(pol, str):
            self._bbox = wkt.loads(pol)
        else:
            self._bbox = pol

    @property
    def camera_matrix(self):
        """Get camera matrix."""
        return self._camera_matrix

    @camera_matrix.setter
    def camera_matrix(self, camera_matrix):
        self._camera_matrix = camera_matrix.tolist() if isinstance(camera_matrix, np.ndarray) else camera_matrix

    @property
    def dist_coeffs(self):
        """Get distortion coefficients."""
        return self._dist_coeffs

    @dist_coeffs.setter
    def dist_coeffs(self, dist_coeffs):
        self._dist_coeffs = dist_coeffs.tolist() if isinstance(dist_coeffs, np.ndarray) else dist_coeffs

    @property
    def focal_length(self):
        """Get focal length."""
        if not self.camera_matrix:
            return None
        return self.camera_matrix[0][0]

    @property
    def k1(self):
        """Get first distortion coefficient."""
        if not self.dist_coeffs:
            return None
        return self.dist_coeffs[0]

    @property
    def k2(self):
        """Get second distortion coefficient."""
        if not self.dist_coeffs:
            return None
        return self.dist_coeffs[1]

    @property
    def gcps_dest(self):
        """Get destination coordinates of GCPs.

        Returns
        -------
        dst : np.ndarray
            destination coordinates of ground control point. z-coordinates are parsed from z_0 if necessary

        """
        if hasattr(self, "gcps"):
            if "dst" in self.gcps:
                return np.array(
                    self.gcps["dst"]
                    if len(self.gcps["dst"][0]) == 3
                    else np.c_[self.gcps["dst"], np.ones(4) * self.gcps["z_0"]],
                    dtype=np.float64,
                )
        # if conditions are not yet met, then return None
        return None

    @property
    def gcps_dest_bbox(self):
        """Give destination coordinates as row, col in intended bounding box.

        Returns
        -------
        dst : np.ndarray
            Destination coordinates measured as column, row in the intended bounding box with the intended resolution

        """
        return np.array(cv.transform_to_bbox(self.gcps_dest, self.bbox, self.resolution))

    @property
    def gcps_bbox_reduced(self):
        """Give col, row coordinates of gcps within intended bounding box, reduced by mean coordinate.

        Returns
        -------
        dst : np.ndarray
            Destination coordinates in col, row in the intended bounding box, reduced with their mean coordinate

        """
        return self.gcps_dest_bbox - self.gcps_dest_bbox.mean(axis=0)

    @property
    def gcps_reduced(self):
        """Get location of gcp destination points, reduced with their mean for better local projection.

        Returns
        -------
        dst : np.ndarray
            Reduced coordinate (x, y) or (x, y, z) of gcp destination points

        """
        return np.array(self.gcps_dest - self.gcps_mean)

    @property
    def gcps_mean(self):
        """Get mean location of gcp destination points.

        Returns
        -------
        dst_mean : np.ndarray
            mean coordinate (x, y) or (x, y, z) of gcp destination points

        """
        return np.array([0.0, 0.0, 0.0]) if self.gcps_dest is None else np.array(self.gcps_dest).mean(axis=0)

    @property
    def gcps_dims(self):
        """Return amount of dimensions of GCPs provided (2 or 3).

        Returns
        -------
        dims : int
            amount of dimensions of gcps (can be 2 or 3)

        """
        return len(self.gcps["dst"][0]) if hasattr(self, "gcps") else None

    @property
    def is_nadir(self):
        """Return if the camera configuration belongs to nadir video.

        Returns
        -------
        is_nadir : bool
            False if not nadir, True if nadir

        """
        return self._is_nadir

    @is_nadir.setter
    def is_nadir(self, nadir_prop: bool):
        self._is_nadir = nadir_prop

    @property
    def pnp(self):
        """Return Precise N point solution from ground control points, intrinsics and distortion."""
        # solve rvec and tvec with reduced coordinates, this ensure that the solvepnp solution is stable.
        _, rvec, tvec = cv.solvepnp(self.gcps_reduced, self.gcps["src"], self.camera_matrix, self.dist_coeffs)
        # ensure that rvec and tvec are corrected for the fact that mean gcp location was subtracted
        rvec_cam, tvec_cam = cv.pose_world_to_camera(rvec, tvec)
        tvec_cam += self.gcps_mean
        # transform back to world
        rvec, tvec = cv.pose_world_to_camera(rvec_cam, tvec_cam)
        return rvec, tvec

    @property
    def rvec(self):
        """Return rvec from precise N point solution."""
        return self.pnp[0].tolist() if self._rvec is None else self._rvec

    @rvec.setter
    def rvec(self, _rvec):
        self._rvec = _rvec.tolist() if isinstance(_rvec, np.ndarray) else _rvec

    @property
    def shape(self):
        """Return rows and columns in projected frames from `Frames.project`.

        Returns
        -------
        rows : int
            Amount of rows in projected frame
        cols : int
            Amount of columns in projected frame

        """
        cols, rows = cv._get_shape(self.bbox, resolution=self.resolution, round=1)
        return rows, cols

    @property
    def stabilize(self):
        """Return stabilization polygon (anything outside is used for stabilization.

        Returns
        -------
        coords : list of lists
            coordinates in original image frame comprising the polygon for use in stabilization

        """
        return self._stabilize

    @stabilize.setter
    def stabilize(self, coords: List[List[float]]):
        self._stabilize = coords

    @property
    def rotation(self):
        """Return rotation OpenCV code.

        Returns
        -------
        code : int
            integer code belonging to rotation, 0 for 90 deg, 1 for 180 deg and 2 for 270 deg

        """
        if hasattr(self, "_rotation"):
            return self._rotation
        else:
            return None

    @rotation.setter
    def rotation(self, rotation_code: int):
        self._rotation = rotation_code

    @property
    def transform(self):
        """Returns Affine transform of projected frames from `Frames.project`.

        Returns
        -------
        transform : rasterio.transform.Affine object

        """
        return cv._get_transform(self.bbox, resolution=self.resolution)

    @property
    def tvec(self):
        """Return tvec from precise N point solution."""
        return self.pnp[1].tolist() if self._tvec is None else self._tvec

    @tvec.setter
    def tvec(self, _tvec):
        self._tvec = _tvec.tolist() if isinstance(_tvec, np.ndarray) else _tvec

    def set_lens_calibration(
        self,
        fn: str,
        chessboard_size: Optional[Tuple] = (9, 6),
        max_imgs: Optional[int] = 30,
        plot: Optional[bool] = True,
        progress_bar: Optional[bool] = True,
        **kwargs,
    ):
        """Calibrate and set the properties `camera_matrix` and `dist_coeffs` using a video of a chessboard pattern.

        Follows methods described on https://docs.opencv.org/4.x/dc/dbb/tutorial_py_calibration.html

        Parameters
        ----------
        fn : str
            path to video file
        df : int, optional
            amount of frames to skip after a valid frame with corner points is found, defaults to fps of video.
        chessboard_size : tuple, optional
            amount of internal corner points expected on chessboard pattern, default is (9, 6).
        max_imgs : int, optional
            maximum amount of images to use for calibration (default: 30).
        tolerance : float, optional
            error tolerance alowed for reprojection of corner points (default: 0.1, if set to None, no filtering will
            be done). images that exceed the tolerance are excluded from calibration. This is to remove images with
            spuriously defined points, or blurry images.
        plot : bool, optional
            if set, chosen frames will be plotted for the user to inspect on-=the-fly with a one-second delay
            (default: True).
        progress_bar : bool, optional
            if set, a progress bar going through the frames is plotted (default: True).
        **kwargs : dict
            keyword arguments to pass to underlying methods

        """
        assert os.path.isfile(fn), f"Video calibration file {fn} not found"
        camera_matrix, dist_coeffs = cv.calibrate_camera(fn, chessboard_size, max_imgs, plot, progress_bar, **kwargs)
        self.camera_matrix = camera_matrix
        self.dist_coeffs = dist_coeffs

    def estimate_lens_position(self):
        """Estimate lens position from distortion and intrinsec/extrinsic matrix."""
        rvec, tvec = np.array(self.rvec), np.array(self.tvec)
        rmat = cv2.Rodrigues(rvec)[0]
        # determine lens position related to center of objective
        lens_pos = (np.array(-rmat).T @ tvec).flatten()
        return lens_pos

    def get_bbox(
        self,
        camera: Optional[bool] = False,
        mode: Optional[MODES] = "geographical",
        h_a: Optional[float] = None,
        z_a: Optional[float] = None,
        within_image: Optional[bool] = False,
        expand_exterior=True,
    ) -> Polygon:
        """Get bounding box.

        Can be retrieved in geographical or camera perpective.

        Parameters
        ----------
        camera : bool, optional
            If set, the bounding box will be returned as row and column coordinates in the camera perspective.
            In this case ``h_a`` may be set to provide the right water level, to estimate the bounding box for.
            This option is deprecated, instead use mode="camera".
        mode : Literal["geographical", "camera", "3d"], optional
            Determines the type of bounding box to return. If set to "geographical" (default), the bounding box
            is returned in the geographical coordinates. If set to "camera", the bounding box is returned in the
            camera perspective. If set to "3d", the bounding box is returned as a 3D polygon in the CRS
        h_a : float, optional
            If set with `mode="camera"`, then the bbox coordinates will be transformed to the camera perspective,
            using h_a as a present water level (in local datum). In case a video with higher (lower) water levels is
            used, this will result in a different perspective plane than the control video.
        z_a : float, optional
            similar to setting h_a, but z_a represent the vertical coordinate in the coordinate reference system of
            the camera configuration instead of the local datum.
        within_image : bool, optional (default False)
            Set to True to make an attempt to remove parts of the polygon that lie outside of the image field of view
        expand_exterior : bool, optional
            Set to True to expand the corner points to more points. This is particularly useful for plotting purposes.

        Returns
        -------
        A bounding box, that in the used CRS is perfectly rectangular, and aligned in the up/downstream direction.
        It can (and certainly will) be rotated with respect to a typical bbox with xlim, ylim coordinates.
        If user sets ``mode="camera"`` then the geographical bounding box will be converted into a camera perspective,
        using the homography belonging to the available ground control points and current water level.

        This can then be used to reconstruct the grid for velocimetry calculations.

        """
        if camera:
            import warnings

            warnings.warn(
                "The camera=True option is deprecated, use mode='camera' instead. This option will be removed in "
                "a future release.",
                DeprecationWarning,
                stacklevel=2,
            )
            mode = "camera"
        bbox = self.bbox
        coords = np.array(bbox.exterior.coords)
        if within_image:
            # in this case, always more points than just corners are needed, so expand_exterior is forced to True
            expand_exterior = True
        if expand_exterior:
            # make a new set of bbox coordinates with a higher density. This is meant to enable plotting of
            # distortion on image frame, and to plot partial coverage in the real-world coordinates
            coords_expand = np.zeros((0, 2))
            for n in range(0, len(coords) - 1):
                new_coords = np.linspace(coords[n], coords[n + 1], 100)
                coords_expand = np.r_[coords_expand, new_coords]
            coords = coords_expand
        if not z_a:
            z_a = self.get_z_a(h_a)
        # add vertical coordinates to the set
        coords = np.c_[coords, np.ones(len(coords)) * z_a]
        # project points to pixel image coordinates
        corners = self.project_points(coords, within_image=within_image)
        corners = corners[np.isfinite(corners[:, 0])]
        if not mode == "camera":
            # project back to real-world coordinates after possibly cutting at edges of visibility
            corners = self.unproject_points(np.array(np.array(list(zip(*corners))).T), z_a)
        if mode == "3d":
            return Polygon(corners[np.isfinite(corners[:, 0])])
        return Polygon(corners[np.isfinite(corners[:, 0])][:, 0:2])

    def get_depth(self, z: List[float], h_a: Optional[float] = None) -> List[float]:
        """Retrieve depth for measured bathymetry points.

        This is done using the camera configuration and an actual water level,
        measured in local reference (e.g. staff gauge).

        Parameters
        ----------
        z : list of floats
            measured bathymetry point depths
        h_a : float, optional
            actual water level measured [m], if not set, assumption is that a single video
            is processed and thus changes in water level are not relevant. (default: None)

        Returns
        -------
        depths : list of floats

        """
        if h_a is None:
            h_a = self.gcps["h_ref"]
        z_pressure = np.maximum(self.gcps["z_0"] - self.gcps["h_ref"] + h_a, z)
        return z_pressure - z

    def get_dist_shore(
        self, x: List[float], y: List[float], z: List[float], h_a: Optional[float] = None
    ) -> List[float]:
        """Retrieve depth for measured bathymetry points.

        This is done using the camera configuration and an actual water level, measured
        in local reference (e.g. staff gauge).

        Parameters
        ----------
        x : list of floats
            measured bathymetry point x-coordinates
        y : list of floats
            measured bathymetry point y-coordinates
        z : list of floats
            measured bathymetry point depths
        h_a : float, optional
            actual water level measured [m], if not set, assumption is that a single video
            is processed and thus changes in water level are not relevant. (default: None)

        Returns
        -------
        depths : list of floats

        """
        # retrieve depth
        depth = self.get_depth(z, h_a=h_a)
        if h_a is None:
            assert self.gcps["h_ref"] is None, (
                "No actual water level is provided, but a reference water level is " "provided "
            )
            # h_a = 0.
            # h_ref = 0.
        # else:
        # h_ref = self.gcps["h_ref"]
        z_dry = depth <= 0
        z_dry[[0, -1]] = True
        # compute distance to nearest dry points with Pythagoras
        dist_shore = np.array([(((x[z_dry] - _x) ** 2 + (y[z_dry] - _y) ** 2) ** 0.5).min() for _x, _y in zip(x, y)])
        return dist_shore

    def get_dist_wall(self, x: List[float], y: List[float], z: List[float], h_a: Optional[float] = None) -> List[float]:
        """Retrieve distance to wall for measured bathymetry points.

        Done by using the camera configuration and an actual water level, measured in local reference (e.g. staff
        gauge).

        Parameters
        ----------
        x : list of floats
            measured bathymetry point x-coordinates
        y : list of floats
            measured bathymetry point y-coordinates
        z : list of floats
            measured bathymetry point depths
        h_a : float, optional
            actual water level measured [m], if not set, assumption is that a single video
            is processed and thus changes in water level are not relevant. (default: None)

        Returns
        -------
        distance : list of floats

        """
        depth = self.get_depth(z, h_a=h_a)
        dist_shore = self.get_dist_shore(x, y, z, h_a=h_a)
        dist_wall = (dist_shore**2 + depth**2) ** 0.5
        return dist_wall

    def get_extrinsic(self):
        """Return rotation and translation vector based on control points and intrinsic parameters."""
        # solve rvec and tvec with reduced coordinates, this ensure that the solvepnp solution is stable.
        _, rvec, tvec = cv.solvepnp(self.gcps_reduced, self.gcps["src"], self.camera_matrix, self.dist_coeffs)
        # ensure that rvec and tvec are corrected for the fact that mean gcp location was subtracted
        rvec_cam, tvec_cam = cv.pose_world_to_camera(rvec, tvec)
        tvec_cam += self.gcps_mean
        # transform back to world
        rvec, tvec = cv.pose_world_to_camera(rvec_cam, tvec_cam)
        return rvec, tvec

    def z_to_h(self, z: float) -> float:
        """Convert z coordinates of bathymetry to height coordinates in local reference (e.g. staff gauge).

        Parameters
        ----------
        z : float
            measured bathymetry point

        Returns
        -------
        h : float

        """
        h_ref = 0 if self.gcps["h_ref"] is None else self.gcps["h_ref"]
        h = z + h_ref - self.gcps["z_0"]
        return h

    def h_to_z(self, h_a: float) -> float:
        """Convert z coordinates of bathymetry to height coordinates in local reference (e.g. staff gauge).

        Parameters
        ----------
        h_a : float
            measured level in local datum

        Returns
        -------
        z : float
            level in global datum

        """
        h_ref = 0 if self.gcps["h_ref"] is None else self.gcps["h_ref"]
        return h_a - h_ref + self.gcps["z_0"]

    def get_M(
        self, h_a: Optional[float] = None, to_bbox_grid: Optional[bool] = False, reverse: Optional[bool] = False
    ) -> np.ndarray:
        """Establish a transformation matrix for a certain actual water level `h_a`.

        This is done by mapping where the ground control points, measured at `h_ref` will end up with new water level
        `h_a`, given the lens position.

        Parameters
        ----------
        h_a : float, optional
            actual water level [m] (Default: None)
        to_bbox_grid : bool, optional
            if set, the M will be computed in row, column values of the target bbox, with set resolution
        reverse : bool, optional
            if True, the reverse matrix is prepared, which can be used to transform projected
            coordinates back to the original camera perspective. (Default: False)

        Returns
        -------
        M : np.ndarray
            2x3 transformation matrix

        """
        src = cv.undistort_points(self.gcps["src"], self.camera_matrix, self.dist_coeffs)
        if to_bbox_grid:
            dst_a = self.gcps_bbox_reduced
        else:
            dst_a = self.gcps_reduced
        # compute the water level in the coordinate system reduced with the mean gcp coordinate
        z_a = self.get_z_a(h_a)
        z_a -= self.gcps_mean[-1]
        # treating 3D homography
        return cv.get_M_3D(
            src=src,
            dst=dst_a,
            camera_matrix=self.camera_matrix,
            dist_coeffs=cv.DIST_COEFFS,  # self.dist_coeffs,
            z=z_a,
            reverse=reverse,
        )

    def get_z_a(self, h_a: Optional[float] = None) -> float:
        """Get actual water level measured in global vertical datum (+z_0) from water level in local datum (+h_ref).

        Parameters
        ----------
        h_a : float, optional
            actual water level measured [m], if not set, assumption is that a single video
            is processed and thus changes in water level are not relevant. (default: None)

        Returns
        -------
        Actual locations of control points (in case these are only x, y) given the current set water level and
        the camera location

        """
        if h_a is None:
            return self.gcps["z_0"]
        else:
            return self.gcps["z_0"] + (h_a - self.gcps["h_ref"])

    def map_idx_img_ortho(self, x, y, z):
        """Map pixel indices from the image to an orthographic projection for nearest resampling.

        Can be used independently, or (when mean resampling is used) with `map_idx_img_ortho_mean`.

        Parameters
        ----------
        x : np.ndarray
            1D array representing the x-coordinates of the target orthographic grid.
        y : np.ndarray
            1D array representing the y-coordinates of the target orthographic grid.
        z : float
            The z-coordinate (elevation) value to compute the transformation.

        Returns
        -------
        idx_img : np.ndarray
            1D array of flattened indices of the source image pixels that correspond
            to selected orthographic grid pixels.
        idx_ortho : np.ndarray
            1D array of flattened indices of the destination ortho image pixels that correspond
            to the source pixels in idx_img.

        """
        # make a large flattened list of coordinates of target grid.
        cols, rows = np.meshgrid(np.arange(len(x)), np.arange(len(y)))
        xs, ys = helpers.pixel_to_map(cols.flatten(), rows.flatten(), self.transform)
        # back-project real-world coordinates to camera coordinates
        points_cam = self.project_points(list(zip(xs, ys, np.ones(len(xs)) * z)))
        # round cam coordinates to pixels
        points_cam = np.int64(np.round(points_cam))
        # find locations that lie within the camera objective, rest should remain missing value
        idx_ortho = np.all(
            [
                points_cam[:, 0] > 0,
                points_cam[:, 0] < self.width,
                points_cam[:, 1] > 0,
                points_cam[:, 1] < self.height,
            ],
            axis=0,
        )
        # check if there are values inside the objective. If not raise a warning
        if idx_ortho.sum() == 0:
            warnings.warn(
                f"The water level is either very low or high compared to the reference water level. "
                f"As a result, there are no pixels in the objective that fit in the area of interest. "
                f"Difference in water level and reference water level is {z - self.gcps['z_0']}. You will get "
                f"missing values only.",
                stacklevel=2,
            )
        # coerce 2D idxs to 1D idxs
        idx_img = np.array(points_cam[idx_ortho, 1]) * self.width + np.array(points_cam[idx_ortho, 0])
        return idx_img, idx_ortho

    def map_mean_idx_img_ortho(self, x, y, z):
        """Map pixel indices from the image to an orthographic projection for mean resampling.

        This function performs a mapping of pixel indices from one coordinate
        system (e.g., camera coordinate system) to another orthographic grid,
        with additional filtering based on specified constraints. It calculates
        the target pixel locations on the orthographic grid, checks which pixels
        reside inside the area of interest, and returns indexes and index groups
        for use in a mean resampling. undersampled areas will not be fully covered,
        hence this method should be used in conjunction with `map_idx_img_ortho`.

        Parameters
        ----------
        x : np.ndarray
            1D array representing the x-coordinates of the target orthographic grid.
        y : np.ndarray
            1D array representing the y-coordinates of the target orthographic grid.
        z : float
            The z-coordinate (elevation) value to compute the transformation.

        Returns
        -------
        src_idx : np.ndarray
            1D array of flattened indices of the source image pixels that correspond
            to selected orthographic grid pixels.
        uidx : np.ndarray
            Sorted unique indices of the filtered orthographic grid pixels.
        norm_idx : np.ndarray
            Normalized indices corresponding to their positions in the unique
            filtered orthographic grid pixels.

        """
        # mapping of pixels from image to ortho
        coli, rowi = np.meshgrid(np.arange(self.width), np.arange(self.height))
        # first exclude pixels not within the area of interest, can save a lot of time
        poly = self.get_bbox(mode="camera", z_a=z)
        mask = rasterize([poly], out_shape=(self.height, self.width)) == 1
        src_pix = list(zip(coli[mask], rowi[mask]))

        # orthoproject pixels
        dst_pix = self.unproject_points(src_pix, z)
        x_pix, y_pix, z_pix = dst_pix.T
        idx_y, idx_x = helpers.map_to_pixel(x_pix, y_pix, self.transform)
        # ensure no pixels outside of target grid (can be in case of edges)
        # idx_inside contains False/True which pixels are within AOI from selected
        idx_inside = np.all([idx_y >= 0, idx_y < len(y), idx_x >= 0, idx_x < len(x)], axis=0)
        idx_x = idx_x[idx_inside]
        idx_y = idx_y[idx_inside]
        # get 1D flat array indexes
        idx = np.array(idx_y) * len(x) + np.array(idx_x)
        src_pix_sel = np.array(src_pix)[idx_inside]

        uidx, counts = np.unique(idx, return_counts=True)

        # Filter out indices that occur only once (covered by nearest-neighbour)
        valid_idx = uidx[counts > 1]
        mask = np.isin(idx, valid_idx)  # Mask to filter data and indices
        src_pix_sel = src_pix_sel[mask]

        # flattened indexes of source img to read
        src_idx = src_pix_sel[:, 1] * self.width + src_pix_sel[:, 0]

        # Apply the mask to both data and indices
        # filtered_data = data[mask]
        filtered_idx = idx[mask]
        # convert into a ascending lists of indexes, which are easier to process in numba
        uidx, norm_idx = np.unique(filtered_idx, return_inverse=True)
        return src_idx, uidx, norm_idx

    def set_bbox_from_corners(self, corners: List[List[float]]):
        """Establish bbox based on a set of camera perspective corner points.

        Parameters
        ----------
        corners : list of lists (4)
            [columns, row] coordinates in original camera perspective without any undistortion applied

        """
        assert np.array(corners).shape == (4, 2), (
            f"a list of lists of 4 coordinates must be given, resulting in (4, "
            f"2) shape. Current shape is {corners.shape} "
        )
        assert self.gcps["z_0"] is not None, "The water level must be set before the bounding box can be established."

        # get homography
        corners_xyz = self.unproject_points(corners, np.ones(4) * self.gcps["z_0"])
        bbox = cv.get_aoi(corners_xyz, resolution=self.resolution)
        self.bbox = bbox

    def set_bbox_from_width_length(self, points: List[List[float]]):
        """Establish bbox based on three provided points.

        The points are provided in the original camera perspective as [col, row] and require that a water level
        has already been set in order to project them in a feasible way.

        first point : left bank (seen in downstream direction)
        second point : right bank
        third point : selected upstream or downstream of the two points.

        The last point defines how large the bounding box is in up-and-downstream direction. A user should attempt to
        choose the first two points roughly in the middle of the intended bounding box. The last point is then
        used to estimate the length perpendicular to the line between the first two points. The bounding box is
        extended in both directions with the same length.

        Parameters
        ----------
        points : list of lists (3)
            [columns, row] coordinates in original camera perspective without any undistortion applied

        """
        assert np.array(points).shape == (3, 2), (
            f"a list of lists of 3 coordinates must be given, resulting in (3, "
            f"2) shape. Current shape is {np.array(points).shape} "
        )
        assert self.gcps["z_0"] is not None, "The water level must be set before the bounding box can be established."
        # get homography
        points_xyz = self.unproject_points(points, np.ones(3) * self.gcps["z_0"])
        bbox = cv.get_aoi(points_xyz, resolution=self.resolution, method="width_length")
        self.bbox = bbox

    def rotate_translate_bbox(self, angle: float = None, xoff: float = None, yoff: float = None):
        """Rotate and translate the bounding box.

        Parameters
        ----------
        angle : float, optional
            Rotation angle in radians (anti-clockwise) around the center of the bounding box
        xoff : float, optional
            Translation distance in x direction in CRS units
        yoff : float, optional
            Translation distance in y direction in CRS units

        Returns
        -------
        CameraConfig
            New CameraConfig instance with rotated and translated bounding box

        """
        # Make a deep copy of current config
        new_config = copy.deepcopy(self)

        # Get the current bbox
        bbox = new_config.bbox
        if bbox is None:
            return new_config

        # Apply rotation if specified
        if angle is not None:
            print(angle)
            # # Convert to radians
            # angle = np.radians(rotation)
            # Get centroid as origin
            centroid = bbox.centroid
            # Apply rotation around centroid
            bbox = shapely.affinity.rotate(
                bbox,
                angle,
                origin=centroid,
                use_radians=True,
            )

        # Now perform translation. Get coordinates of corners
        coords = list(bbox.exterior.coords)

        # Get unit vectors of x and y directions
        p1 = np.array(coords[0])
        p2 = np.array(coords[1])  # second point
        p3 = np.array(coords[2])  # third point

        x_vec = p2 - p1
        y_vec = p3 - p2

        x_vec = x_vec / np.linalg.norm(x_vec)
        y_vec = y_vec / np.linalg.norm(y_vec)
        # Project translations onto these vectors
        dx = 0 if xoff is None else xoff * x_vec[0]
        dy = 0 if xoff is None else xoff * x_vec[1]

        dx -= 0 if yoff is None else yoff * y_vec[0]
        dy -= 0 if yoff is None else yoff * y_vec[1]

        # Apply translation
        bbox = shapely.affinity.translate(bbox, xoff=dx, yoff=dy)
        new_config.bbox = bbox
        return new_config

    def calibrate(
        self,
    ):
        """Calibrate camera parameters using ground control.

        If nothing provided, they are derived by optimizing pnp fitting together with optimizing the focal length
        and two radial distortion coefficients (k1, k2).

        You may also provide camera matrix or distortion coefficients, which will only optimize
        the remainder parameters.

        As a result, the following is updated on the CameraConfig instance:
        - camera_matrix: the 3x3 camera matrix
        - dist_coeffs: the 5x1 distortion coefficients
        - rvec: the 3x1 rotation vector
        - tvec: the 3x1 translation vector
        """
        if hasattr(self, "gcps") and (self.camera_matrix is None or self.dist_coeffs is None):
            # some calibration is needed, and there are GCPs available for it
            if len(self.gcps["src"]) >= 4:
                self.camera_matrix, self.dist_coeffs, err = cv.optimize_intrinsic(
                    self.gcps["src"],
                    self.gcps_dest,
                    # self.gcps["dst"],
                    self.height,
                    self.width,
                    lens_position=self.lens_position,
                    camera_matrix=self.camera_matrix,
                    dist_coeffs=self.dist_coeffs,
                )
        # finally, also derive the rvec and tvec if camera matrix and distortion coefficients are known
        if self.camera_matrix is not None and self.dist_coeffs is not None:
            rvec, tvec = self.get_extrinsic()
            self.rvec = rvec
            self.tvec = tvec

    def set_gcps(
        self, src: List[List], dst: List[List], z_0: float, h_ref: Optional[float] = None, crs: Optional[Any] = None
    ):
        """Set ground control points for the given CameraConfig.

        Parameters
        ----------
        src : list of lists (2, 4 or 6+)
            [x, y] pairs of columns and rows in the frames of the original video
        dst : list of lists (2, 4 or 6+)
            [x, y] or [x, y, z] pairs of real world coordinates in the given coordinate reference system.
        z_0 : float
            Water level measured in global reference system such as a geoid or ellipsoid used
            by a GPS device. All other surveyed points (lens position and cross section) must have the same vertical
            reference.
        h_ref :  float, optional
            Water level, belonging to the 4 control points in `dst`. This is the water level
            as measured by a local reference (e.g. gauge plate) during the surveying of the control points. Control
            points must be taken on the water surface. If a single movie is processed, h_ref can be left out.
            (Default: None)
        crs : int, dict or str, optional
            Coordinate Reference System. Accepts EPSG codes (int or str) proj (str or dict) or wkt (str). CRS used to
            measure the control points (e.g. 4326 for WGS84 lat-lon). Destination control points will automatically be
            reprojected to the local crs of the CameraConfig. (Default: None)

        """
        assert isinstance(src, list), "src must be a list of (x, y) or (x, y, z) coordinates"
        assert isinstance(dst, list), "dst must be a list of (x, y) or (x, y, z) coordinates"
        if np.array(dst).shape[1] == 2:
            assert len(src) in [2, 4], f"2 or 4 source points are expected in src, but {len(src)} were found"
            if len(src) == 4:
                assert len(dst) == 4, f"4 destination points are expected in dst, but {len(dst)} were found"
            else:
                assert len(dst) == 2, f"2 destination points are expected in dst, but {len(dst)} were found"
        else:
            assert len(src) == len(
                dst
            ), f"Amount of (x, y, z) coordinates in src ({len(src)}) and dst ({len(dst)} must be equal"
            assert len(src) >= 6, f"for (x, y, z) points, at least 6 pairs must be available, only {len(src)} provided"
        if h_ref is not None:
            assert isinstance(h_ref, (float, int)), "h_ref must contain a float number"
        if z_0 is not None:
            assert isinstance(z_0, (float, int)), "z_0 must be provided as type float"
        assert all(isinstance(x, (float, int)) for p in src for x in p), "src contains non-int parts"
        assert all(isinstance(x, (float, int)) for p in dst for x in p), "dst contains non-float parts"
        if crs is not None:
            if not (hasattr(self, "crs")):
                raise ValueError(
                    "CameraConfig does not contain a crs, so gcps also cannot contain a crs. Ensure that the provided "
                    "destination coordinates are in a locally defined coordinate reference system, e.g. established "
                    "with a spirit level."
                )
            dst = helpers.xyz_transform(dst, crs, CRS.from_wkt(self.crs))
        # if there is no h_ref, then no local gauge system, so set h_ref to zero
        # check if 2 points are available
        if len(src) == 2:
            self.is_nadir = True
            src, dst = cv._get_gcps_2_4(src, dst, self.width, self.height)
        if h_ref is None:
            h_ref = 0.0
        self.gcps = {
            "src": src,
            "dst": dst,
            "h_ref": h_ref,
            "z_0": z_0,
        }

    def set_lens_position(self, x: float, y: float, z: float, crs: Optional[Any] = None):
        """Set the geographical position of the lens of current CameraConfig.

        Parameters
        ----------
        x : float
            x-coordinate
        y : float
            y-coordinate
        z : float
            z-coordinate
        crs : int, dict or str, optional
            Coordinate Reference System. Accepts EPSG codes (int or str) proj (str or dict) or wkt (str). CRS used to
            measure the lens position (e.g. 4326 for WGS84 lat-lon). The position's x and y coordinates will
            automatically be reprojected to the local crs of the CameraConfig.

        """
        if crs is not None:
            if self.crs is None:
                raise ValueError("CameraConfig does not contain a crs, ")
            x, y = helpers.xyz_transform([[x, y]], crs, self.crs)[0]
        self.lens_position = [x, y, z]

    def project_points(self, points: List[List], within_image=False, swap_y_coords=False) -> np.ndarray:
        """Project real world x, y, z coordinates into col, row coordinates on image.

        If col, row coordinates are not allowed to go outside of the image frame, then set `within_image = True`.
        Method uses the intrinsics and extrinsics and distortion parameters to perform the projection.

        Parameters
        ----------
        points : list of lists or array-like
            list of points [x, y, z] in real world coordinates
        within_image : bool, optional
            Set coordinates to NaN if these fall outside of the image.
        swap_y_coords : bool, optional
            If set to True (default: False), y-coordinates will be swapped, in order to match plotting defaults
            which return row counting from top to bottom instead of bottom to top.

        Returns
        -------
        points_project : list or array-like
            list of points (equal in length as points) with [col, row] coordinates

        """
        rvec, tvec = np.array(self.rvec), np.array(self.tvec)
        # normalize points wrt mean of gcps
        points = np.array(points, dtype=np.float64)
        points_proj, jacobian = cv2.projectPoints(
            points, rvec, tvec, np.array(self.camera_matrix), np.array(self.dist_coeffs)
        )
        points_proj = np.array([list(point[0]) for point in points_proj])

        # points_back = cv.unproject_points(src=points_proj, z=points[:, -1], )
        if within_image:
            # also filter points outside edges of image
            points_proj[points_proj[:, 0] < 0, 0] = -1.0
            points_proj[points_proj[:, 0] > self.width - 1, 0] = self.width
            points_proj[points_proj[:, 1] < 0, 1] = -1.0
            points_proj[points_proj[:, 1] > self.height - 1, 1] = self.height
            # points_proj[points_proj[:, 0] < 0, 0] = np.nan
            # points_proj[points_proj[:, 0] > self.width - 1, 0] = np.nan  # self.width
            # points_proj[points_proj[:, 1] < 0, 1] = np.nan  # -1.0
            # points_proj[points_proj[:, 1] > self.height - 1, 1] = np.nan  # self.height

            # check which points lie behind the camera
            R, _ = cv2.Rodrigues(rvec)
            points_camera = cv.world_to_camera(points, rvec, tvec)
            behind_camera = points_camera[:, 2] <= 0.0
            # set points behind camera to nan
            points_proj[behind_camera, :] = np.nan
        # swap y coords if set
        if swap_y_coords:
            points_proj[:, 1] = self.height - points_proj[:, 1]
        return points_proj

    def project_grid(self, xs, ys, zs, swap_y_coords=False):
        """Project gridded coordinates to col, row coordinates on image.

        Method uses the intrinsics and extrinsics and distortion parameters to perform the projection.

        Parameters
        ----------
        xs : np.ndarray
            2d array of real-world x-coordinates
        ys : np.ndarray
            2d array of real-world y-coordinates
        zs : np.ndarray
            2d array of real-world z-coordinates
        swap_y_coords : bool, optional
            If set to True (default: False), y-coordinates will be swapped, in order to match plotting defaults
            which return row counting from top to bottom instead of bottom to top.

        Returns
        -------
        xp : np.ndarray
            list of col coordinates of image objective
        yp : np.ndarray
            list of row coordinates of image objective

        """
        points = list(zip(xs.flatten(), ys.flatten(), zs.flatten()))
        points_proj = np.array(self.project_points(points, swap_y_coords=swap_y_coords))
        xp, yp = points_proj[:, 0], points_proj[:, 1]
        # reshape back
        xp = np.reshape(xp, (len(xs), -1))
        yp = np.reshape(yp, (len(xs), -1))
        return xp, yp

    def unproject_points(self, points: List[List], zs: Union[float, List[float]]) -> np.ndarray:
        """Reverse projects points in [column, row] space to [x, y, z] real world.

        Parameters
        ----------
        points : List of lists or array-like
            Points in [col, row] to unproject
        zs : float or list of floats
            z-coordinates on which to unproject points

        Returns
        -------
        points_unproject : List of lists or array-like
            unprojected points as list of [x, y, z] coordinates

        """
        rvec, tvec = np.array(self.rvec), np.array(self.tvec)
        # reduce zs by the mean of the gcps
        dst = cv.unproject_points(
            np.array(points, dtype=np.float64),
            zs,
            rvec=rvec,
            tvec=tvec,
            camera_matrix=self.camera_matrix,
            dist_coeffs=self.dist_coeffs,
        )
        dst = np.array(dst, dtype=np.float64)
        return dst

    def plot(
        self,
        figsize: Optional[Tuple] = (13, 8),
        ax: Optional[plt.Axes] = None,
        tiles: Optional[Any] = None,
        buffer: Optional[float] = 0.0005,
        zoom_level: Optional[int] = 19,
        camera: Optional[bool] = False,
        mode: Optional[MODES] = "geographical",
        pose_length: float = 1.0,
        tiles_kwargs: Optional[Dict] = None,
    ) -> plt.Axes:
        """Plot geographical situation of the CameraConfig.

        This is very useful to check if the CameraConfig seems to be in the right location. Requires `cartopy`
        to be installed.

        Parameters
        ----------
        figsize : tuple, optional
            width and height of figure (Default value = (13)
        ax : plt.axes, optional
            if not provided, axes is setup (Default: None)
        tiles : str, optional
            name of tiler service to use (called as attribute from cartopy.io.img_tiles) (Default: None)
        buffer : float, optional
            buffer in lat-lon around points, used to set extent (default: 0.0005)
        zoom_level : int, optional
            zoom level of image tiler service (default: 18)
        camera : bool, optional
            If set to True, all camera config information will be back projected to the original camera objective.
            This option is deprecated, instead use mode="camera".
        mode : Literal["geographical", "camera", "3d"], optional
            Determines the type of bounding box to return. If set to "geographical" (default), the bounding box
            is returned in the geographical coordinates. If set to "camera", the bounding box is returned in the
            camera perspective. If set to "3d", the bounding box is returned as a 3D polygon in the CRS
        pose_length: float, optional
            length of pose axes to draw (only used in mode="3d").
        tiles_kwargs : dict
            additional keyword arguments to pass to ax.add_image when tiles are added

        Returns
        -------
        ax : plt.axes

        """
        if camera:
            import warnings

            warnings.warn(
                "The camera=True option is deprecated, use mode='camera' instead. This option will be removed in "
                "a future release.",
                DeprecationWarning,
                stacklevel=2,
            )
            mode = "camera"
        if not tiles_kwargs:
            tiles_kwargs = {}
        # initiate transform
        transformer = None
        # if there is an axes, get the extent
        xlim = ax.get_xlim() if ax is not None else None
        ylim = ax.get_ylim() if ax is not None else None

        # prepare points for plotting
        if mode == "camera":
            points = [Point(x, y) for x, y in self.gcps["src"]]
        elif mode == "geographical":
            points = [Point(p[0], p[1]) for p in self.gcps["dst"]]
        else:
            # 3d points are needed
            if hasattr(self, "gcps"):
                if len(self.gcps["dst"]) == 3:
                    points = [Point(*p) for p in self.gcps["dst"]]
                else:
                    points = [Point(p[0], p[1], self.gcps["z_0"]) for p in self.gcps["dst"]]
        if mode != "camera":
            if self.lens_position is not None:
                lens_position = self.lens_position
            else:
                lens_position = self.estimate_lens_position()
            if mode == "3d":
                points.append(Point(*lens_position))
            else:
                points.append(Point(lens_position[0], lens_position[1]))
            # transform points in case a crs is provided and we want a geographical plot
            if mode == "geographical" and hasattr(self, "crs"):
                # make a transformer to lat lon
                transformer = Transformer.from_crs(
                    CRS.from_user_input(self.crs), CRS.from_epsg(4326), always_xy=True
                ).transform
                points = [ops.transform(transformer, p) for p in points]
            if mode == "geographical":
                xmin, ymin, xmax, ymax = list(np.array(LineString(points).bounds))
                extent = [xmin - buffer, xmax + buffer, ymin - buffer, ymax + buffer]
        x = [p.x for p in points]
        y = [p.y for p in points]
        if mode == "3d":
            z = [p.z for p in points]
        if ax is None:
            plt.figure(figsize=figsize)
            if hasattr(self, "crs") and mode == "geographical":
                ax = helpers.get_geo_axes(tiles=tiles, extent=extent, zoom_level=zoom_level, **tiles_kwargs)
            else:
                if mode == "3d":
                    ax = plt.axes(projection="3d")
                else:
                    ax = plt.axes()
        if hasattr(ax, "add_geometries"):
            import cartopy.crs as ccrs

            plot_kwargs = dict(transform=ccrs.PlateCarree())
        else:
            plot_kwargs = {}
        if hasattr(self, "gcps"):
            if mode == "3d":
                ax.plot(
                    x[0 : len(self.gcps["dst"])],
                    y[0 : len(self.gcps["dst"])],
                    z[0 : len(self.gcps["dst"])],
                    "o",
                    label="Control points",
                    markersize=12,
                    markeredgecolor="w",
                    zorder=2,
                    **plot_kwargs,
                )
            else:
                ax.plot(
                    x[0 : len(self.gcps["dst"])],
                    y[0 : len(self.gcps["dst"])],
                    ".",
                    label="Control points",
                    markersize=12,
                    markeredgecolor="w",
                    zorder=2,
                    **plot_kwargs,
                )
        # if len(x) > len(self.gcps["dst"]):
        if mode == "3d":
            ax.plot(
                x[-1],
                y[-1],
                z[-1],
                "o",
                label="Lens position",
                markersize=12,
                zorder=2,
                markeredgecolor="w",
                **plot_kwargs,
            )
            # add pose
            _ = self.plot_3d_pose(ax=ax, length=pose_length)
            if hasattr(self, "bbox"):
                # also plot dashed lines from cam to bbox
                for xy in self.bbox.exterior.coords:
                    ax.plot([x[-1], xy[0]], [y[-1], xy[1]], [z[-1], self.gcps["z_0"]], linestyle="--", color="gray")
                # plot bbox exterior
                ax.plot(*self.bbox.exterior.xy, [self.gcps["z_0"]] * 5, color="k", label="bbox exterior")
        else:
            ax.plot(
                x[-1],
                y[-1],
                ".",
                label="Lens position",
                markersize=12,
                zorder=2,
                markeredgecolor="w",
                **plot_kwargs,
            )
        patch_kwargs = {
            **plot_kwargs,
            "alpha": 0.5,
            "zorder": 2,
            "edgecolor": "w",
            "label": "bbox visible",
            **plot_kwargs,
        }
        if hasattr(self, "bbox"):
            self.plot_bbox(ax=ax, mode=mode, transformer=transformer, within_image=True, **patch_kwargs)
        if mode == "camera":
            # make sure that zero is on the top
            ax.set_aspect("equal")
            if xlim is not None:
                ax.set_xlim(xlim)
                ax.set_ylim(ylim)
            ax.set_xlabel("column [-]")
            ax.set_ylabel("row [-]")
        elif mode == "3d":
            ax.set_xlabel("x [m]")
            ax.set_ylabel("y [m]")
            ax.set_zlabel("z [m]")
        ax.legend()
        return ax

    def plot_bbox(
        self,
        ax: Optional[plt.Axes] = None,
        camera: Optional[bool] = False,
        mode: Optional[MODES] = "geographical",
        transformer: Optional[Any] = None,
        h_a: Optional[float] = None,
        within_image: Optional[bool] = True,
        **kwargs,
    ):
        """Plot bounding box.

        This can be done for orthorectification in a geographical projection (`camera=False`) or the camera
        Field Of View (`mode="camera"`).

        Parameters
        ----------
        ax : plt.axes, optional
            if not provided, axes is setup (Default: None)
        camera : bool, optional
            If set to True, all camera config information will be back projected to the original camera objective.
            This option is deprecated, instead use mode="camera".
        mode : Literal["geographical", "camera", "3d"], optional
            Determines the type of bounding box to return. If set to "geographical" (default), the bounding box
            is returned in the geographical coordinates. If set to "camera", the bounding box is returned in the
            camera perspective. If set to "3d", the bounding box is returned as a 3D polygon in the CRS
        transformer : pyproj transformer transformation function, optional
            used to reproject bbox to axes object projection (e.g. lat lon)
        h_a : float, optional
            If set with `mode="camera"`, then the bbox coordinates will be transformed to the camera perspective,
            using h_a as a present water level. In case a video with higher (lower) water levels is used, this
            will result in a different perspective plane than the control video.
        within_image : bool, optional
            If set (default), points outside the camera objective are removed.
        **kwargs : dict
            additional keyword arguments used for plotting the bbox polygon with `matplotlib.patches.Polygon`

        Returns
        -------
        p : matplotlib.patch mappable

        """
        # collect information to plot
        if camera:
            import warnings

            warnings.warn(
                "The camera=True option is deprecated, use mode='camera' instead. This option will be removed in "
                "a future release.",
                DeprecationWarning,
                stacklevel=2,
            )
            mode = "camera"
        bbox = self.get_bbox(mode=mode, h_a=h_a, within_image=within_image)
        if mode == "geographical" and transformer is not None:
            # geographical projection is needed
            bbox = ops.transform(transformer, bbox)
        if mode == "3d":
            return plot_helpers.plot_3d_polygon(bbox, ax=ax, **kwargs)
        return plot_helpers.plot_polygon(bbox, ax=ax, **kwargs)

    def plot_3d_pose(self, ax=None, length=1):
        """Plot 3D pose of a camera using its rotation and translation vectors.

        Parameters
        ----------
        ax : axes, optional
            3d axes to plot on, if not set, a new axes will be established.
        length : float
            length of the axes drawn in meters

        Returns
        -------
        list[handles]
            list of handles to the plotted pose axes

        """
        rvec = np.array(self.rvec)
        tvec = np.array(self.tvec)
        # rvec, tvec = cv.pose_world_to_camera(rvec, tvec)
        # Convert the rotation vector to a 3x3 rotation matrix
        R, _ = cv2.Rodrigues(rvec.flatten())

        # Define the camera's axis directions in its local coordinate system
        camera_axes = (
            np.array(
                [
                    [0, 0, 0],  # lens center
                    [1, 0, 0],  # X-axis (red - right looking)
                    [0, 1, 0],  # Y-axis (green - down looking)
                    [0, 0, 1],  # Z-axis (blue - forward looking)
                ]
            )
            * length
        )
        pts_trans = camera_axes - tvec
        world_axes_translated = (R.T @ pts_trans.T).T
        ax = plt.axes(projection="3d") if ax is None else ax
        # Plot the origin of the camera
        ps = []
        # Plot the camera axes
        for i, (color, label) in enumerate(zip(["r", "g", "b"], ["right-pose", "down-pose", "forward-pose"])):
            # if i == 2:
            xx = [world_axes_translated[0, 0], world_axes_translated[i + 1, 0]]
            yy = [world_axes_translated[0, 1], world_axes_translated[i + 1, 1]]
            zz = [world_axes_translated[0, 2], world_axes_translated[i + 1, 2]]
            ps.append(ax.plot(xx, yy, zz, color=color, label=label, linewidth=3))
        return ps

    def to_dict(self) -> Dict:
        """Return the CameraConfig object as dictionary.

        Returns
        -------
        camera_config_dict : dict
            serialized CameraConfig

        """
        d = copy.deepcopy(self.__dict__)
        # replace underscore keys for keys without underscore
        for k in list(d.keys()):
            if k[0] == "_":
                d[k[1:]] = d.pop(k)

        return d

    def to_dict_str(self) -> Dict:
        """Convert the current instance to a dictionary with all values converted to strings.

        Returns
        -------
        dict
            A dictionary representation of the instance where all values are strings. If an attribute is an instance
            of `Polygon`, it is converted to its string representation.

        """
        d = self.to_dict()
        # convert anything that is not string in string
        dict_str = {k: v if not (isinstance(v, Polygon)) else v.__str__() for k, v in d.items()}
        return dict_str

    def to_file(self, fn: str):
        """Write the CameraConfig object to json structure.

        Parameters
        ----------
        fn : str
            Path to file to write camera config to

        """
        with open(fn, "w") as f:
            f.write(self.to_json())

    def to_json(self) -> str:
        """Convert CameraConfig object to string.

        Returns
        -------
        json_str : str
            json string with CameraConfig components

        """
        return json.dumps(self, default=lambda o: o.to_dict_str(), indent=4)


depr_warning_height_width = """
Your camera configuration does not have a property "height" and/or "width", probably because your configuration file is
from an older < 0.3.0 version. Please rectify this by editing your .json config file. The top of your file should e.g.
look as follows for a HD video:
{
    "height": 1080,
    "width": 1920,
    "crs": ....
    ...
}
"""


def get_camera_config(s: str) -> CameraConfig:
    """Read camera config from string.

    Parameters
    ----------
    s : str
        json string containing camera config

    Returns
    -------
    cam_config : CameraConfig

    """
    d = json.loads(s)
    if "height" not in d or "width" not in d:
        raise IOError(depr_warning_height_width)
    # ensure the bbox is a Polygon object
    if "bbox" in d:
        if isinstance(d["bbox"], str):
            d["bbox"] = wkt.loads(d["bbox"])
    return CameraConfig(**d)


def load_camera_config(fn: str) -> CameraConfig:
    """Load a CameraConfig from a geojson file.

    Parameters
    ----------
    fn : str
        path to file with camera config in json format.

    Returns
    -------
    cam_config : CameraConfig

    """
    with open(fn, "r") as f:
        camera_config = get_camera_config(f.read())
    return camera_config



"""Logging module for pyorc CLI."""

import logging
import os
import sys

from pyorc import __version__

FMT = "%(asctime)s - %(name)s - %(module)s - %(levelname)s - %(message)s"


class CustomFormatter(logging.Formatter):
    """Adapted formatter for pyorc."""

    grey = "\x1b[38;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    cyan = "\x1b[36;20m"
    reset = "\x1b[0m"
    format = FMT

    FORMATS = {
        logging.DEBUG: cyan + format + reset,
        logging.INFO: grey + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset,
    }

    def format(self, record):
        """Get format conditional on record level."""
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def setuplog(
    name: str = "pyorc",
    path: str = None,
    log_level: int = 20,
    fmt: str = FMT,
    append: bool = True,
) -> logging.Logger:
    """Set up the logging on sys.stdout and file if path is given.

    Parameters
    ----------
    name : str, optional
        logger name, by default "hydromt"
    path : str, optional
        path to logfile, by default None
    log_level : int, optional
        Log level [0-50], by default 20 (info)
    fmt : str, optional
        log message formatter, by default {FMT}
    append : bool, optional
        Whether to append (True) or overwrite (False) to a logfile at path, by default True

    Returns
    -------
    logging.Logger
        _description_

    """
    logger = logging.getLogger(name)
    for _ in range(len(logger.handlers)):
        logger.handlers.pop().close()  # remove and close existing handlers
    logging.captureWarnings(True)
    logger.setLevel(log_level)
    console = logging.StreamHandler(sys.stdout)
    console.setLevel(log_level)
    # console.setFormatter(logging.Formatter(fmt))
    console.setFormatter(CustomFormatter())
    logger.addHandler(console)
    if path is not None:
        if append is False and os.path.isfile(path):
            os.unlink(path)
        add_filehandler(logger, path, log_level=log_level, fmt=fmt)
    logger.info(f"pyorc version: {__version__}")

    return logger


def add_filehandler(logger, path, log_level=20, fmt=FMT):
    """Add file handler to logger."""
    if not os.path.isdir(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))
    isfile = os.path.isfile(path)
    ch = logging.FileHandler(path)
    ch.setFormatter(logging.Formatter(fmt))
    ch.setLevel(log_level)
    logger.addHandler(ch)
    if isfile:
        logger.debug(f"Appending log messages to file {path}.")
    else:
        logger.debug(f"Writing log messages to new file {path}.")



"""Helper functions for pyorc."""

import copy
import importlib.util
import json

import cv2
import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from pyproj import Transformer
from rasterio import fill, warp
from rasterio.crs import CRS
from rasterio.transform import Affine, xy
from scipy.interpolate import interp1d
from scipy.ndimage import binary_fill_holes
from scipy.optimize import differential_evolution
from scipy.signal import convolve2d, fftconvolve


def _check_cartopy_installed():
    if importlib.util.find_spec("cartopy") is None:
        raise ModuleNotFoundError(
            'Geographic plotting requires cartopy. Please install it with "conda install cartopy" and try again.'
        )


def _import_cartopy_modules():
    import cartopy.crs as ccrs
    import cartopy.io.img_tiles as cimgt

    return ccrs, cimgt


def affine_from_grid(xi, yi):
    """Retrieve the affine transformation from a gridded set of coordinates.

    This function (unlike rasterio.transform functions) can also handle rotated grids

    Parameters
    ----------
    xi: np.ndarray (2D)
        gridded x-coordinates
    yi: np.ndarray (2D)
        gridded y-coordinates

    Returns
    -------
    obj : rasterio.transform.Affine

    """
    xul, yul = xi[0, 0], yi[0, 0]
    xcol, ycol = xi[0, 1], yi[0, 1]
    xrow, yrow = xi[1, 0], yi[1, 0]
    dx_col = xcol - xul
    dy_col = ycol - yul
    dx_row = xrow - xul
    dy_row = yrow - yul
    return Affine(dx_col, dy_col, xul, dx_row, dy_row, yul)


def densify_points(points, sample_size=1000):
    """Increase the amount of points through linear interpolation. points are assumed to be sorted in space.

    Parameters
    ----------
    points : np.ndarray
        at minimum one-dimensional array with point coordinates

    sample_size : int
        amount of samples to expand to by linear interpolation

    """
    idx = np.arange(len(points))
    # generate interpolation function
    f = interp1d(np.arange(len(points)), points, axis=0)
    # expand to amount of samples
    return f(np.linspace(0, idx.max(), sample_size))


def depth_integrate(depth, v, v_corr=0.85, name="q"):
    """Integrate velocities [m s-1] to depth-integrated velocity [m2 s-1] using depth information.

    Parameters
    ----------
    depth : DataArray (points)
        bathymetry depths (ref. CRS)
    v : DataArray (time, points)
        effective velocity at surface [m s-1]
    v_corr : float (range: 0-1), optional
        typically close to 1, correction factor from surface to depth-average (default: 0.85)
    name: str, optional
        name of DataArray (default: "q")

    Returns
    -------
    q: DataArray (time, points)
        depth integrated velocity [m2 s-1]

    """
    # compute the depth average velocity
    q = v * v_corr * depth
    q.attrs = {
        "standard_name": "velocity_depth",
        "long_name": "velocity averaged over depth",
        "units": "m2 s-1",
    }
    # set name
    q.name = name
    return q


def deserialize_attr(data_array, attr, dtype=np.array, args_parse=False):
    """Return a deserialized version of said property (assumed to be stored as a string) of DataArray.

    Parameters
    ----------
    data_array : xr.DataArray
        attributes of interest
    attr : str
        name of attributes
    dtype : object type, optional
        function will try to perform type(eval(attr)), default np.array
    args_parse : boolean, optional
        if True, function will try to return type(*eval(attr)), assuming attribute contains list
        of arguments (default: False)

    Returns
    -------
    parsed_arg : type defined by user
        parsed attribute, of type defined by arg type

    """
    assert hasattr(data_array, attr), f'frames do not contain attribute with name "{attr}'
    attr_obj = getattr(data_array, attr)
    if args_parse:
        return dtype(*eval(attr_obj))
    return dtype(eval(attr_obj))


def get_axes(cols, rows, x, y):
    """Retrieve a locally spaced axes for surface velocimetry results.

    axes are based on resolution and row and col distances from the original frames

    Parameters
    ----------
    cols: list
        ints, columns, sampled from the original projected frames
    rows: list
        ints, rows, sampled from the original projected frames
    x: array-like
        frames x-axis
    y: array-like
        frames y-axis

    Returns
    -------
    xax : np.ndarray
        x-axis sampled from columns
    yax : np.ndarray
        y-axis sampled from columns

    """
    xax = x[cols]
    yax = y[rows]
    return xax, yax


def get_geo_axes(tiles=None, extent=None, zoom_level=19, **kwargs):
    """Prepare a geographical axis, possibly with an image tiler background.

    Parameters
    ----------
    tiles : str
        Name of cartopy.io.img_tiles tiler
    extent : List[int]
        [xmin, xmax, ymin, ymax] extent in longitude, latitude coordinates
    zoom_level : int
        zoom level to use for image tiler, increase if you want higher resolution imagery
    **kwargs : dict
        additional keyword arguments to pass to the chosen image tiler.

    Returns
    -------
    ax
        Geographical cartopy axis object

    """
    _check_cartopy_installed()
    #
    # try:
    #     import cartopy
    #     import cartopy.crs as ccrs
    #     import cartopy.io.img_tiles as cimgt
    # except ModuleNotFoundError:
    #     raise ModuleNotFoundError(
    #         'Geographic plotting requires cartopy. Please install it with "conda install cartopy" and try ' "again."
    #     )
    ccrs, cimgt = _import_cartopy_modules()
    if tiles is not None:
        tiler = getattr(cimgt, tiles)(**kwargs)
        crs = tiler.crs
    else:
        tiler = None
        crs = ccrs.PlateCarree()
    ax = plt.subplot(projection=crs)
    if extent is not None:
        ax.set_extent(extent, crs=ccrs.PlateCarree())
    if tiles is not None:
        ax.add_image(tiler, zoom_level, zorder=1)
    return ax


def get_enclosed_mask(data, stride=2):
    """Create zero/one binary mask that can be used to fill holes.

    The function fills holes in yes/no finite data in a numpy array with data and nans and
    only fills where NaN areas are entirely enclosed by finite values.

    Parameters
    ----------
    data : 2D numpy array
        dataset for which mask should be created
    stride : int, optional
        amount of cells that will be parsed at the edges to ensure holes at edges are also filled.

    Returns
    -------
    mask : 2D numpy array
        zero/one array containing which areas are expected to contain data after filling procedures


    """
    mask = np.zeros(data.shape)
    mask[np.isfinite(data)] = 1

    # explode mask by one pixel in all directions
    mask_edge = np.minimum(fftconvolve(mask, np.ones((stride * 2 + 1, stride * 2 + 1))), 1)

    # mask_edge = np.minimum(convolve2d(np.ones((stride*2 + 1, stride*2 + 1)), mask), 1)
    mask_edge[stride:-stride, stride:-stride] = mask

    # fill holes of areas that are entirely enclosed
    mask_edge_fill = binary_fill_holes(mask_edge)

    mask_filled = np.int8(mask_edge_fill[stride:-stride, stride:-stride])
    mask_filled[mask_filled == 0] = -1
    mask_filled[mask_filled == 1 & np.isnan(data)] = 0
    return mask_filled


def get_rotation_code(rotation_code):
    """Convert rotation in degrees into a opencv rotation code.

    Parameters
    ----------
    rotation_code : Literal[0, 90, 180, 270]
        degrees rotation of a video or image

    Returns
    -------
    int
        opencv rotation code

    """
    if rotation_code not in [0, 90, 180, 270, None]:
        raise ValueError(f"Rotation code must be in allowed codes 0, 90, 180 or 270. Provided code is {rotation_code}")
    if rotation_code == 90:
        return cv2.ROTATE_90_CLOCKWISE
    elif rotation_code == 180:
        return cv2.ROTATE_180
    elif rotation_code == 270:
        return cv2.ROTATE_90_COUNTERCLOCKWISE
    else:
        return None


def get_xs_ys(cols, rows, transform):
    """Compute rasters of x and y coordinates, based on row and column counts and a defined transform.

    Parameters
    ----------
    cols: list of ints
        column counts
    rows: list of ints
        row counts
    transform: np.ndarray (1D)
        rasterio compatible transform parameters

    Returns
    -------
    xs : np.ndarray (MxN)
        x-coordinates
    ys : np.ndarray (MxN)
        y-coordinates

    """
    xs, ys = xy(transform, rows, cols)
    xs, ys = np.array(xs), np.array(ys)
    return xs, ys


def get_lons_lats(xs, ys, src_crs, dst_crs=4326):
    """Create lon-lat rasters from local projection coordinate grid.

    Computes raster of longitude and latitude coordinates (default) of a certain raster set of coordinates in a local
    coordinate reference system. User can supply an alternative coordinate reference system if projection other than
    WGS84 Lat Lon is needed.

    Parameters
    ----------
    xs : np.ndarray (MxN)
        x-coordinates
    ys : np.ndarray (MxN)
        y-coordinates
    src_crs : int, dict or str
        Coordinate Reference System (of source coordinates). Accepts EPSG codes (int or str) proj (str or dict) or wkt
        (str).
    dst_crs : int, dict or str, optional
        Coordinate Reference System (of target coordinates). Accepts EPSG codes (int or str) proj (str or dict) or wkt
        (str). default: CRS.from_epsg(4326) for wGS84 lat-lon

    Returns
    -------
    lons : np.ndarray (MxN)
        longitude coordinates
    lats: np.ndarray (MxN)
        latitude coordinates

    """
    dst_crs = CRS.from_user_input(dst_crs)
    lons, lats = warp.transform(src_crs, dst_crs, xs.flatten(), ys.flatten())
    lons, lats = (
        np.array(lons).reshape(xs.shape),
        np.array(lats).reshape(ys.shape),
    )
    return lons, lats


def log_profile(x, z0, k_max, s0=0.0, s1=0.0):
    """Return values of a log-profile function.

    Parameters
    ----------
    x: tuple with np.ndarrays
        (depth [m], distance to bank [m]) arrays of equal length
    z0: float
        depth with zero velocity [m]
    k_max: float
        maximum scale factor of log-profile function [-]
    s0: float, optional
        distance from bank (default: 0.) where k equals zero (and thus velocity is zero) [m]
    s1: float, optional
        distance from bank (default: 0. meaning no difference over distance) where k=k_max (k cannot be larger than
        k_max) [m]

    Returns
    -------
    velocity : np.ndarray
        V values from log-profile, equal shape as arrays inside X [m s-1]

    """
    z, s = x
    k = k_max * np.minimum(np.maximum((s - s0) / (s1 - s0), 0), 1)
    v = k * np.maximum(np.log(np.maximum(z, 1e-6) / z0), 0)
    return v


def pixel_to_map(cols, rows, transform):
    """Replace `transform.xy` in numpy using `rasterio.transform` order. This is much faster than rasterio.

    Parameters
    ----------
    cols : np.ndarray
        columns set
    rows : np.ndarray
        rows set
    transform : array-like or transform object
        transform of raster

    Returns
    -------
    xs : array-like
        x-coordinates
    ys : array-like
        y-coordinates

    """
    # Affine transformation (assuming transform is the raster's transform)
    x_map = transform[2] + rows * transform[1] + cols * transform[0]
    y_map = transform[5] + rows * transform[4] + cols * transform[3]

    return x_map, y_map


def map_to_pixel(xs, ys, transform):
    """Transform with `numpy` similar to `transform.rowcol` using `rasterio.transform`.

    Numpy is much faster than rasterio.

    Parameters
    ----------
    xs : np.ndarray
        x-coordinates
    ys : np.ndarray
        y-coordinates
    transform : array-like or transform object
        transform of raster

    Returns
    -------
    rows : array-like
        row coordinates
    cols : array-like
        column coordinates

    """
    # Calculate the determinant of the upper-left 2x2 submatrix
    det = transform[1] * transform[3] - transform[0] * transform[4]

    # Calculate the inverse of the upper-left 2x2 submatrix
    inv_det = 1.0 / det
    inv_transform = [transform[3] * inv_det, -transform[0] * inv_det, -transform[4] * inv_det, transform[1] * inv_det]

    # Calculate the offsets
    dx = xs - transform[2]
    dy = ys - transform[5]

    # Calculate the pixel coordinates
    row = np.int64(np.round(inv_transform[0] * dx + inv_transform[1] * dy))
    col = np.int64(np.round(inv_transform[2] * dx + inv_transform[3] * dy))

    return row, col


def mask_fill(data, mask, radius=5):
    """Fill data where np.nan is found, if mask at those location is zero.

    Areas where mask is one are used to fill up these areas. Areas where mask is -1 are not filled and kept np.nan

    Parameters
    ----------
    data : np.ndarray (2D)
        data values with possibly np.nan values to fill
    mask : np.ndarray (2D, type int8)
        mask values to apply, either -1, 0 or 1
    radius : int, float
        search distance radius [pix]


    Returns
    -------
    data_fill : np.ndarray (2D)
        data filled

    """
    mask[np.isfinite(data)] = 1
    data_fill = copy.deepcopy(fill.fillnodata(data, mask=mask == 1, max_search_distance=radius))
    data_fill[mask == -1] = 0
    return data_fill


def mse(pars, func, x, y):
    """Give mean of sum of squares between evaluation of function with provided parameters and observations.

    Parameters
    ----------
    pars : list or tuple
        parameter passed as *args to func.
    func: function def
        receiving X and *pars as input and returning predicted Y as result
    x: tuple with lists or array-likes
        indepent variable(s).
    y: list or array-like
        dependent variable, predicted by func

    Returns
    -------
    y_pred : list or array-like
        predicted Y from X and pars by func

    """
    y_pred = func(x, *pars)
    ms_error = np.sum((y_pred - y) ** 2)
    return ms_error


def neighbour_stack(array, stride=1, missing=-9999.0):
    """Build stack of arrays from a 2-D input array, constructed by permutation in space using a provided stride.

    Parameters
    ----------
    array : np.ndarray (2D)
        any values (may contain NaN)
    stride : int, optional
        stride used to determine relevant neighbours (default: 1)
    missing : float, optional
        a temporary missing value, used to be able to convolve NaNs

    Returns
    -------
    obj : np.array (3D)
        stack of 2-D arrays, with strided neighbours (length 1st dim : (stride*2+1)**2 )

    """
    array = copy.deepcopy(array)
    array[np.isnan(array)] = missing
    array_move = []
    for vert in range(-stride, stride + 1):
        for horz in range(-stride, stride + 1):
            conv_arr = np.zeros((abs(vert) * 2 + 1, abs(horz) * 2 + 1))
            _y = int(np.floor((abs(vert) * 2 + 1) / 2)) + vert
            _x = int(np.floor((abs(horz) * 2 + 1) / 2)) + horz
            conv_arr[_y, _x] = 1
            array_move.append(convolve2d(array, conv_arr, mode="same", fillvalue=np.nan))
    array_move = np.stack(array_move)
    # replace missings by Nan
    array_move[np.isclose(array_move, missing)] = np.nan
    return array_move


def optimize_log_profile(
    z,
    v,
    dist_bank=None,
    bounds=([0.001, 0.1], [-20, 20], [0.0, 5], [0.0, 100]),
    workers=2,
    popsize=100,
    updating="deferred",
    seed=0,
    **kwargs,
):
    """Optimize velocity log profile relation of v=k*max(z/z0) with k a function of distance to bank and k_max.

    A differential evolution optimizer is used.

    Parameters
    ----------
    z : list
        depths [m]
    v : list
        surface velocities [m s-1]
    dist_bank : list, optional
        distances to bank [m]
    bounds : tuple of int-pairs, optional
        search boundaries for the optimization function, default: ([0.001, 0.1], [-20, 20], [0.0, 5], [0.0, 100])
    workers : int, optional
        Amount of workers used to optimize log profile function (default=2) in `scipy.optimize.differential_evolution`
    popsize : int, optional
        size of search population (default: 100) in `scipy.optimize.differential_evolution`
    updating : str, optional
        method of updating in `scipy.optimize.differential_evolution` (default: "deferred").
    seed : float, optional
        seed point used in `scipy.optimize.differential_evolution` (default: 0).
    **kwargs : dict
        additional keyword arguments for scipy.optimize.differential_evolution

    Returns
    -------
    pars : dict
        fitted parameters of log_profile {z_0, k_max, s0 and s1}

    """
    # replace by infinites if not provided
    dist_bank = np.ones(len(v)) * np.inf if dist_bank is None else dist_bank
    v = np.array(v)
    z = np.array(z)
    x = (z, dist_bank)
    y = v
    result = differential_evolution(
        wrap_mse,
        args=(log_profile, x, y),
        bounds=bounds,
        workers=workers,
        popsize=popsize,
        updating=updating,
        seed=seed,
        **kwargs,
    )
    # unravel parameters
    z0, k_max, s0, s1 = result.x
    return {"z0": z0, "k_max": k_max, "s0": s0, "s1": s1}


def read_shape_safe_crs(fn):
    """Read a shapefile with geopandas, but ensure that CRS is set to None when not available.

    This function is required in cases where geometries must be read that do not have a specified CRS. Geopandas
    defaults to WGS84 EPSG 4326 if the CRS is not specified.
    """
    gdf = gpd.read_file(fn)
    # also read raw json, and check if crs attribute exists
    if isinstance(fn, str):
        with open(fn, "r") as f:
            raw_json = json.load(f)
    else:
        # apparently a file object was provided
        fn.seek(0)
        raw_json = json.load(fn)
    if "crs" not in raw_json:
        # override the crs
        gdf = gdf.set_crs(None, allow_override=True)
    return gdf


def rotate_u_v(u, v, theta, deg=False):
    """Rotate u and v components of vector counterclockwise by an amount of rotation.

    Parameters
    ----------
    u : float, np.ndarray or xr.DataArray
        x-direction component of vector
    v : float, np.ndarray or xr.DataArray
        y-direction component of vector
    theta : float
        amount of counterclockwise rotation in radians or degrees (dependent on deg)
    deg : boolean, optional
        if True, theta is defined in degrees, otherwise radians (default: False)

    Returns
    -------
    u_rot : float, np.ndarray or xr.DataArray
        rotated x-direction component of vector
    v_rot : float, np.ndarray or xr.DataArray
        rotated y-direction component of vector

    """
    theta = np.radians(theta) if deg else theta
    c, s = np.cos(theta), np.sin(theta)
    r = np.array(((c, -s), (s, c)))
    # compute rotations with dot-product
    u2 = r[0, 0] * u + r[0, 1] * v
    v2 = r[1, 0] * u + r[1, 1] * v
    return u2, v2


def round_to_multiple(number, multiple):
    """Round number to a multiple of a certain number."""
    return multiple * round(number / multiple)


def stack_window(ds, wdw=1, wdw_x_min=None, wdw_x_max=None, wdw_y_min=None, wdw_y_max=None, dim="stride"):
    """Stack moved windows together over a new dimension.

    Typically used to average or take a quantile over a window of pixels in space

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to stack over
    wdw : int (positive), optional
        general all-directional stride to use (default: 1)
    wdw_x_min : int, optional
        Stride to use in negative x-axis direction, if not provided, -`wdw` is used.
    wdw_x_max
        Stride to use in positive x-axis direction, if not provided, `wdw` is used.
    wdw_y_min
        Stride to use in negative y-axis direction, if not provided, -`wdw` is used.
    wdw_y_max
        Stride to use in positive y-axis direction, if not provided, `wdw` is used.
    dim : str, optional
        Name of new stack dimension

    Returns
    -------
    xr.Dataset
        Containing all variables of `ds`, but then window shifted over x and y dimension and stacked.

    """
    # set strides
    wdw_x_min = -wdw if wdw_x_min is None else wdw_x_min
    wdw_x_max = wdw if wdw_x_max is None else wdw_x_max
    wdw_y_min = -wdw if wdw_y_min is None else wdw_y_min
    wdw_y_max = wdw if wdw_y_max is None else wdw_y_max

    return xr.concat(
        [
            ds.shift(x=x_stride, y=y_stride)
            for x_stride in range(wdw_x_min, wdw_x_max + 1)
            for y_stride in range(wdw_y_min, wdw_y_max)
        ],
        dim=dim,
    )


def staggered_index(start=0, end=100):
    """Create staggered indexes that start at the outer indexes and gradually move inwards.

    Parameters
    ----------
    start : int, optional
        start index number (default: 0)
    end : int, optional
        end index number (default: 100)

    Returns
    -------
    idx : list
        staggered indexes from start to end

    """
    # make list of frames in order to read, starting with start + end frame
    idx_order = [start, end]
    # make sorted representation of frames
    idx_sort = np.array(idx_order)
    idx_sort.sort()
    while True:
        idx_new = (np.round((idx_sort[0:-1] + idx_sort[1:]) / 2)).astype("int")
        # check which of these are already on the list
        idx_new = list(set(idx_new).difference(idx_order))
        if len(idx_new) == 0:
            # we have treated all idxs
            break
        idx_order += idx_new
        idx_sort = np.array(idx_order)
        idx_sort.sort()
    return idx_order


def velocity_log_fit(v, depth, dist_shore, dim="quantile"):
    """Fill missing surface velocities using a velocity log-depth model, fitted with the known points in profile.

    Parameters
    ----------
    v : xr.DataArray (time, points)
        effective velocity at surface [m s-1]
    depth : xr.DataArray (points)
        bathymetry depths [m]
    dist_shore : xr.DataArray (points)
        the shortest distance to a dry river bed point
    dim: str, optional
        dimension over which data should be grouped, default: "quantile", dimension must exist in v, typically
        "quantile" or "time"

    Returns
    -------
    v_fill: xr.DataArray (quantile or time, points)
        effective surface velocities with interpolated values added  [m s-1]

    """

    def log_fit(_v):
        idx_finite = np.isfinite(_v).values.flatten()
        pars = optimize_log_profile(depth[idx_finite], _v[0, idx_finite], dist_shore[idx_finite])
        idx_miss = np.where(np.isnan(_v[0]).values)[0]
        _v[0, idx_miss] = log_profile((depth[idx_miss], dist_shore[idx_miss]), **pars)
        # enforce that velocities are zero with zero depth
        _v[0, depth <= 0] = 0.0
        return np.maximum(_v, 0)

    # fill per grouped dimension
    v.load()
    v_group = copy.deepcopy(v).groupby(dim, squeeze=False)
    return v_group.map(log_fit)


def velocity_log_interp(v, dist_wall, d_0=0.1, dim="quantile"):
    """Interpolate missing velocities over log-transform with depth.

    Parameters
    ----------
    v : xr.DataArray (time, points) or (quantile, points)
        effective velocity at surface [m s-1]
    dist_wall : xr.DataArray (points)
        the shortest distance to the river bed
    d_0 : float, optional
        roughness length (default: 0.1)
    dim: str, optional
        dimension over which data should be grouped, default: "quantile", dimension must exist in v, typically
        "quantile" or "time"

    Returns
    -------
    xr.DataArray (time, points) or (quantile, points)
        effective velocities with interpolated values added [m s-1]

    """

    def log_interp(_v):
        # scale with log depth
        c = xr.DataArray(_v / np.log(np.maximum(dist_wall, d_0) / d_0))
        # fill dry points with the nearest valid value for c
        c[0, np.where(dist_wall == 0)[0]] = c.interpolate_na(dim="points", method="nearest", fill_value="extrapolate")[
            0, np.where(dist_wall == 0)[0]
        ]
        # interpolate with linear interpolation
        c = c.interpolate_na(dim="points")
        # use filled c to interpret missing v
        _v[0, np.isnan(_v[0])] = (np.log(np.maximum(dist_wall, d_0) / d_0) * c)[
            0, np.where(np.isnan(_v[0]))[0]
        ]  # (np.log(np.maximum(dist_wall, d_0) / d_0) * c)[np.isnan(_v)]
        return _v

    # fill per grouped dimension
    v.load()
    v_group = copy.deepcopy(v).groupby(dim, squeeze=False)
    return v_group.map(log_interp)


def wrap_mse(pars_iter, *args):
    """Wrap function mse for optimization purposes."""
    return mse(pars_iter, *args)


def xy_equidistant(x, y, distance, z=None):
    """Transform ordered in space x, y (and z if provided) coordinates into equal spaced x, y (and z) coordinates.

    The 1-dimensional distance between points is used with piece-wise linear interpolation. Extrapolation is used for
    the last point to ensure the range of points covers at least the full range of x, y coordinates.

    Parameters
    ----------
    x : np.ndarray (1D)
        set of (assumed ordered) x-coordinates
    y : np.ndarray (1D)
        set of (assumed ordered) x-coordinates
    distance : float
        user demanded distance between equidistant samples measured in cumulated 1-dimensional distance from xy
        origin (first point)
    z : np.ndarray (1D), optional
        set of (assumed ordered) z-coordinates (default: None, meaning only x, y interpolated points are returned)

    Returns
    -------
    x_sample : np.ndarray (1D)
        interpolated x-coordinates for x, y, s (distance from first point), z
    y_sample : np.ndarray (1D)
        interpolated y-coordinates
    s_sample : np.ndarray (1D)
        interpolated s-coordinates, s being piece-wise linear distance from first point
    z_sample : np.ndarray (1D), optional
        interpolated z-coordinates (only returned if z is not None):

    """
    # estimate cumulative distance between points, starting with zero
    x_diff = np.concatenate((np.array([0]), np.diff(x)))
    y_diff = np.concatenate((np.array([0]), np.diff(y)))
    s = np.cumsum((x_diff**2 + y_diff**2) ** 0.5)

    # create interpolation functions for x and y coordinates
    f_x = interp1d(s, x, fill_value="extrapolate")
    f_y = interp1d(s, y, fill_value="extrapolate")

    # make equidistant samples
    s_sample = np.arange(s.min(), np.ceil((1 + s.max() / distance) * distance), distance)

    # interpolate x and y coordinates
    x_sample = f_x(s_sample)
    y_sample = f_y(s_sample)
    if z is None:
        return x_sample, y_sample, s_sample
    else:
        f_z = interp1d(s, z, fill_value="extrapolate")
        z_sample = f_z(s_sample)
        return x_sample, y_sample, z_sample, s_sample


def xy_angle(x, y):
    """Determine angle between x, y points.

    Parameters
    ----------
    x : np.ndarray (1D)
        set of (assumed ordered) x-coordinates
    y : np.ndarray (1D)
        set of (assumed ordered) x-coordinates

    Returns
    -------
    angle : np.ndarray (1D)
        angle between the point left and right of the point under consideration. The most left and right coordinates
        are based on the first and last 2 points respectively

    """
    angles = np.zeros(len(x))
    angles[1:-1] = np.arctan2(x[2:] - x[0:-2], y[2:] - y[0:-2])
    angles[0] = np.arctan2(x[1] - x[0], y[1] - y[0])
    angles[-1] = np.arctan2(x[-1] - x[-2], y[-1] - y[-2])
    return angles


def xy_to_perspective(x, y, resolution, trans_mat, reverse_y=None):
    """Back transform local meters-from-top-left coordinates from frame to original perspective of camera.

    Back transform is done using M matrix, belonging to transformation from orthographic to local.

    Parameters
    ----------
    x : np.ndarray (1D)
        axis of x-coordinates in local projection with origin top-left, to be backwards projected
    y : np.ndarray (1D)
        axis of y-coordinates in local projection with origin top-left, to be backwards projected
    resolution : float
        resolution of original projected frames coordinates of x and y
    trans_mat : np.ndarray
        2x3 transformation matrix (generated with cv2.getPerspectiveTransform)
    reverse_y : int, optional
        if set, it should be set to the amount of rows in the frame, then y will be transformed following
        y = reverse_y - y

    Returns
    -------
    xp : np.ndarray (2D)
        perspective columns with shape len(y), len(x)
    yp : np.ndarray (2D)
        perspective rows with shape len(y), len(x)

    """
    cols, rows = x / resolution - 0.5, y / resolution - 0.5
    if reverse_y is not None:
        rows = reverse_y - rows
    # make list of coordinates, compatible with cv2.perspectiveTransform
    coords = np.float32([np.array([cols.flatten(), rows.flatten()]).transpose([1, 0])])
    coords_trans = cv2.perspectiveTransform(coords, trans_mat)
    xp = coords_trans[0][:, 0].reshape(cols.shape)
    yp = coords_trans[0][:, 1].reshape(cols.shape)
    return xp, yp


def xyz_transform(points, crs_from, crs_to):
    """Transform set of x and y coordinates from one CRS to another.

    Parameters
    ----------
    points : list of lists
        xyz-coordinates or xy-coordinates in crs_from
    crs_from : int, dict or str, optional
        Coordinate Reference System (source). Accepts EPSG codes (int or str) proj (str or dict) or wkt (str).
    crs_to : int, dict or str, optional
        Coordinate Reference System (destination). Accepts EPSG codes (int or str) proj (str or dict) or wkt (str).

    Returns
    -------
    x_trans : np.ndarray
        x-coordinates transformed
    y_trans : np.ndarray
        y-coordinates transformed

    """
    points = np.array(points)
    x = points[:, 0]
    y = points[:, 1]

    transform = Transformer.from_crs(crs_from, crs_to, always_xy=True)
    # with only one point, transformer must not provide an array of points, to prevent a deprecation warning numpy>=1.25
    if len(points) == 1:
        x = x[0]
        y = y[0]
    # transform dst coordinates to local projection
    x_trans, y_trans = transform.transform(x, y)
    # check if finites are found, if not raise error
    assert not (
        np.all(np.isinf(x_trans))
    ), "Transformation did not give valid results, please check if the provided crs of input coordinates is correct."
    points[:, 0] = np.atleast_1d(x_trans)
    points[:, 1] = np.atleast_1d(y_trans)
    return points.tolist()
    # return transform.transform(x, y)
